from __future__ import annotations

from typing import Optional

import cvxpy as cp
import numpy as np
from numpy import ndarray as Array
from scipy.spatial.distance import cdist

from vqr import VQRSolver, VQRSolution
from vqr.vqr import vector_quantile_levels

SIMILARITY_FN_INNER_PROD = lambda x, y: np.dot(x, y)


class CVXVQRSolver(VQRSolver):
    """
    Solves the Optimal Transport formulation of Vector Quantile Regression using
    CVXPY as a solver backend.

    See:
        Carlier, Chernozhukov, Galichon. Vector quantile regression:
        An optimal transport approach,
        Annals of Statistics, 2016
    """

    @classmethod
    def solver_name(cls) -> str:
        return "cvx_primal"

    def __init__(self, verbose: bool = False, **cvx_solver_opts):
        super().__init__()
        self._verbose = verbose
        cvx_solver_opts["verbose"] = verbose
        self._solver_opts = cvx_solver_opts

    def solve_vqr(self, T: int, Y: Array, X: Optional[Array] = None) -> VQRSolution:
        N = len(Y)
        Y = np.reshape(Y, (N, -1))

        ones = np.ones(shape=(N, 1))
        if X is None:
            X = ones
        else:
            X = np.reshape(X, (N, -1))
            X = np.concatenate([ones, X], axis=1)

        k: int = X.shape[1] - 1  # Number of features (can be zero)
        d: int = Y.shape[1]  # number or target dimensions

        X_bar = np.mean(X, axis=0, keepdims=True)  # (1, k+1)

        # All quantile levels
        Td: int = T**d
        U: Array = vector_quantile_levels(T, d)
        assert U.shape == (Td, d)

        # Pairwise distances (similarity)
        S: Array = cdist(U, Y, SIMILARITY_FN_INNER_PROD)  # (Td, d) and (N, d)

        # Optimization problem definition: optimal transport formulation
        one_N = np.ones([N, 1])
        one_T = np.ones([Td, 1])
        Pi = cp.Variable(shape=(Td, N))
        Pi_S = cp.sum(cp.multiply(Pi, S))
        constraints = [
            Pi @ X == 1 / Td * one_T @ X_bar,
            Pi >= 0,
            one_T.T @ Pi == 1 / N * one_N.T,
        ]
        problem = cp.Problem(objective=cp.Maximize(Pi_S), constraints=constraints)

        # Solve the problem
        problem.solve(**self._solver_opts)

        # Obtain the lagrange multipliers Alpha (A) and Beta (B)
        AB: Array = constraints[0].dual_value
        AB = np.reshape(AB, newshape=[Td, k + 1])
        A = AB[:, [0]]  # A is (T**d, 1)
        if k == 0:
            B = None
        else:
            B = AB[:, 1:]  # B is (T**d, k)

        return VQRSolution(T, d, U, A, B)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple, Union, Callable, Optional, Sequence
from itertools import repeat

import ot
import numpy as np
from numpy import ndarray as Array
from sklearn.utils import check_array


class VQRSolution:
    """
    Encapsulates the solution to a VQR problem. Contains the vector quantiles and
    regression coefficients of the solution, and provides useful methods for
    interacting with them.

    Should only be constructed by :class:`VQRSolver`s, not manually.

    Given a sample x of shape (1, k), the conditional d-dimensional vector quantiles
    Y|X=x are given by d/du [ B @ x.T  + A ].
    We first calculate the regression targets Y_hat = B @ x.T  + A.
    This is an array of shape (T**d, 1).
    In order to obtain the d vector quantile surfaces (one for each dimension of Y),
    use the :obj:`decode_quantile_values` function on Y_hat.
    These surfaces can be visualized over the grid defined in U.
    The combination of these surfaces comprise the conditional quantile function,
    Q_{Y|X=x}(u) which has a d-dimensional input and output.
    """

    def __init__(
        self,
        T: int,
        d: int,
        U: Array,
        A: Array,
        B: Optional[Array] = None,
        X_transform: Optional[Callable[[Array], Array]] = None,
        k_in: Optional[int] = None,
        solution_metrics: Optional[Dict[str, Any]] = None,
    ):
        """
        :param U: Array of shape (T**d, d). Contains the d-dimensional grid on
        which the vector quantiles are defined. If can be decoded back into a
        meshgrid-style sequence of arrays using :obj:`decode_quantile_grid`
        :param A: Array of shape (T**d, 1). Contains the  the regression intercept
        variable. This can be decoded into the d vector quantiles of Y using the
        :obj:`decode_quantile_values` function.
        :param B: Array of shape (T**d, k). Contains the  regression coefficients.
        Will be None if the input was an estimation problem (X=None) instead of a
        regression problem. The k is the dimension of the covariates (X). If an
        X_transform is provided, k corresponds to the dimension AFTER the
        transformation.
        :param X_transform: Transformation to apply to covariates (X) for non-linear
        VQR. Must be provided together with k_in. The transformation is assumed to
        take input of shape (N, k_in) and return output of shape (N, k).
        :param k_in: Covariates input dimension of the X_transform.
        If X_transform is None, must be None or zero.
        :param solution_metrics: Optional key-value pairs which can contain any
        metric values which are tracked by the solver, such as losses, runtimes, etc.
        The keys in this dict are solver-specific and should be specified in the
        documentation of the corresponding solver.
        """
        # Validate dimensions
        assert all(x is not None for x in [T, d, U, A])
        assert U.ndim == 2 and A.ndim == 2
        assert U.shape[0] == A.shape[0]
        assert U.shape[0] == T**d
        assert A.shape[1] == 1
        assert B is None or (B.ndim == 2 and B.shape[0] == T**d)
        assert (X_transform is not None and k_in) or (X_transform is None and not k_in)
        assert solution_metrics is None or isinstance(solution_metrics, dict)

        self._T = T
        self._d = d
        self._U = U
        self._A = A
        self._B = B
        self._k = B.shape[1] if B is not None else 0
        self._X_transform = X_transform
        self._k_in = k_in
        self._solution_metrics = solution_metrics or {}

    @property
    def is_conditional(self) -> bool:
        return self._B is not None

    def vector_quantiles(
        self, X: Optional[Array] = None, refine: bool = False
    ) -> Sequence[QuantileFunction]:
        """
        :param X: Covariates, of shape (N, k). Should be None if the fitted solution
        was for a VQE (un conditional quantiles).
        :param refine: Refine the conditional quantile function using vector monotone
        rearrangement.
        :return: A sequence of length N containing QuantileFunction instances
        corresponding to the given covariates X. If X is None, will be a sequence of
        length one.
        """

        if not self.is_conditional:
            if X is not None:
                raise ValueError(f"VQE was fitted but covariates were supplied")

            Y_hats = [self._A]
        else:
            if X is None:
                raise ValueError(f"VQR was fitted but no covariates were supplied")

            check_array(X, ensure_2d=True, allow_nd=False)

            Z = X  # Z represents the transformed X
            if self._X_transform is not None:
                N, k_in = X.shape
                if k_in != self._k_in:
                    raise ValueError(
                        f"VQR model was trained with X_transform expecting k_in"
                        f"={self._k_in}, but got covariates with {k_in=} features."
                    )

                Z = self._X_transform(X)

            N, k = Z.shape
            if k != self._k:
                raise ValueError(
                    f"VQR model was fitted with k={self._k}, "
                    f"but got data with {k=} features."
                )

            B = self._B  # (T**d, k)
            A = self._A  # (T**d, 1) -> will be broadcast to (T**d, N)
            Y_hat = B @ Z.T + A  # result is (T**d, N)
            Y_hats = Y_hat.T  # (N, T**d)

        refine_fn = lambda Qs: (
            vector_monotone_rearrangement(self._T, self._d, Qs) if refine else Qs
        )
        return tuple(
            QuantileFunction(
                T=self._T,
                d=self._d,
                Qs=refine_fn(decode_quantile_values(self._T, self._d, Y_hat)),
                Us=decode_quantile_grid(self._T, self._d, self._U),
                X=X[[i], :] if X is not None else None,
            )
            for i, Y_hat in enumerate(Y_hats)
        )

    @property
    def quantile_grid(self) -> Sequence[Array]:
        """
        :return: A sequence of quantile level grids as ndarrays. This is a
            d-dimensional meshgrid (see np.meshgrid) where d is the dimension of the
            target variable Y.
        """
        return decode_quantile_grid(self._T, self._d, self._U)

    @property
    def quantile_levels(self) -> Array:
        """
        :return: An array containing the levels at which the vector quantiles were
            estimated along each target dimension.
        """
        return quantile_levels(self._T)

    @property
    def dim_y(self) -> int:
        """
        :return: The dimension d, of the target variable (Y).
        """
        return self._d

    @property
    def dim_x(self) -> int:
        """
        :return: The dimension k, of the covariates (X) which are expected to be
        passed in to obtain conditional quantiles.
        """
        # If there was an X_transform, the input dimension of that is what we expect to
        # receive.
        return self._k_in or self._k

    @property
    def metrics(self):
        """
        :return: A dict containing solver-specific metrics about the solution.
        """
        return self._solution_metrics.copy()


class QuantileFunction:
    """
    Represents a discretized conditional vector-valued quantile function Q_{Y|X=x}(u)
    of the variable Y|X, where:

    - Y is a d-dimensional target variable
    - X is a k-dimensional covariates vector
    - u is a d-dimensional quantile level.
    - Q_{Y|X=x}(u) is the d-dimensional quantile of Y|X=x at level u.

    This instances of this class expose both the d quantile surfaces of the function,
    and also allow evaluating the function at a given quantile level u.

    - Iterating over an instance or indexing it yields the d quantile surfaces,
      e.g. list(q) or q[i].
    - Using an instance as a function evaluates the quantile function, eg. q(u).
    """

    def __init__(
        self,
        T: int,
        d: int,
        Qs: Sequence[Array],
        Us: Sequence[Array],
        X: Optional[Array] = None,
    ):
        """
        :param T: The number of quantile levels (in each dimension).
        :param d: The dimension of the target variable.
        :param Qs: Quantile surfaces of the quantile function: d arrays,
        each d-dimensional with shape (T, T, ..., T).
        :param Us: Quantile levels per dimension of Y. A sequence of length d,
        where each element is of shape (T, T, ..., T).
        :param X: The covariates on which this quantile function is conditioned.
        """
        if len(Qs) != d:
            raise ValueError(
                f"Expecting {d=} quantile surfaces in quantile function, got {len(Qs)}"
            )

        if not all(Q.shape == tuple([T] * d) for Q in Qs):
            raise ValueError(
                f"Expecting {T=} levels in each dimension of each quantile surface"
            )

        if not (len(Us) == len(Qs) and all(U.shape == Q.shape for Q, U in zip(Qs, Us))):
            raise ValueError(f"Expecting Us and Qs to match in number and shape")

        if X is not None:
            if np.ndim(X) > 2 or (np.ndim(X) == 2 and X.shape[0] > 1):
                raise ValueError(
                    f"Unexpected shape of X, must be (1, k), got {X.shape}"
                )

            X = np.reshape(X, (1, -1))

        self._Qs = np.stack(Qs, axis=0)  # Stack into shape (d,T,T,...,T)
        assert self._Qs.shape == tuple([d, *[T] * d])

        self._Us = np.stack(Us, axis=0)  # Stack into shape (d,T,T,...,T)
        assert self._Us.shape == self._Qs.shape

        self.T = T
        self.d = d
        self.X = X
        self.k = 0 if X is None else X.shape[1]

    @property
    def values(self) -> Array:
        """
        :return: All the discrete values of this quantile function.
        A (d+1)-dimensional array of shape (d, T, T, ... T), where the first axis
        indexes different quantile surfaces.
        """
        return self._Qs

    @property
    def levels(self) -> Array:
        """
        :return: All the discrete quantile levels of this quantile function.
        A (d+1)-dimensional array of shape (d, T, T, ... T), where the first axis
        indexes different quantile surfaces.
        """
        return self._Us

    def __iter__(self):
        """
        :return: An iterator over the quantile surfaces (Qs) of this quantile function.
        Yields d elements, each a d-dimensional array of shape (T, T, ..., T).
        """
        for d_idx in range(self.d):
            yield self._Qs[d_idx]

    def __len__(self):
        """
        :return: Number of dimensions of this quantile function.
        """
        return self.d

    def __getitem__(self, d_idx: int) -> Array:
        """
        :param d_idx: An index of a dimension of the target variable, in range [0, d).
        :return: Quantile surface for that dimension.
        """
        return self._Qs[d_idx]

    def __call__(self, u: Array) -> Array:
        """
        :param u: d-dimensional quantile level represented as the integer index of
        the level in each dimension. Each entry of u must be an integer in [0, T-1].
        For example if d=2, T=10 then u=[3, 9] represents the quantile level [0.4, 1.0]
        :return: The vector-quantile value at level u, i.e. Q_{Y|X=x}(u).
        """
        u = u.astype(np.int)

        if not np.ndim(u) == 1 or not len(u) == self.d:
            raise ValueError(f"u must be of shape (d,), got shape {u.shape}")

        if not np.all((u >= 0) & (u < self.T)):
            raise ValueError(
                f"u must contain indices of quantile levels, each in [0, T-1]"
            )

        q_idx = (slice(None), *u)
        q = self._Qs[q_idx]
        assert q.shape == (self.d,)
        return q


class VQRSolver(ABC):
    """
    Abstraction of a method for solving the Vector Quantile Regression (VQR) problem.

    For an overview of the VQR problem and it's solutions, refer to:
        Carlier G, Chernozhukov V, De Bie G, Galichon A.
        Vector quantile regression and optimal transport, from theory to numerics.
        Empirical Econometrics, 2020.
    """

    @classmethod
    @abstractmethod
    def solver_name(cls) -> str:
        """
        :return: An identifier for this solver.
        """
        pass

    @abstractmethod
    def solve_vqr(
        self,
        T: int,
        Y: Array,
        X: Optional[Array] = None,
    ) -> VQRSolution:
        """
        Solves the provided VQR problem in an implementation-specific way.

        :param T: Number of quantile levels to estimate along each of the d
        dimensions. The quantile level will be spaced uniformly between 0 to 1.
        :param Y: The regression target variable, of shape (N, d) where N is the
        number of samples and d is the dimension of the target, which is also the
        dimension of the quantiles which will be estimated.
        :param X: The regression input (features, covariates), of shape (N, k),
        where k is the number of features. Note that X may be None, in which case the
        problem becomes quantile estimation (estimating the quantiles of Y) instead
        of quantile regression.
        :return: A VQRSolution containing the vector quantiles and regression coefficients.
        """
        pass


def quantile_levels(T: int) -> Array:
    """
    Creates a vector of evenly-spaced quantile levels between zero and one.
    :param T: Number of levels to create.
    :return: An array of shape (T,).
    """
    return (np.arange(T) + 1) * (1 / T)


def vector_quantile_levels(T: int, d: int) -> Array:
    """
    Constructs an array of vector-quantile levels. Each level is a d-dimensional
    vector with entries in [0,1].

    The quantile levels are returned stacked along the first axis of the result.
    The output of this function can be converted into a "meshgrid" with
    decode_quantile_grid().

    :param T: Number of levels in each dimension.
    :param d: Dimension of target variable.
    :return: A (T^d, d) array with a different level at each "row".
    """
    # Number of all quantile levels
    Td: int = T**d

    # Quantile levels grid: list of grid coordinate matrices, one per dimension
    # d arrays of shape (T,..., T)
    U_grids: Sequence[Array] = np.meshgrid(*([quantile_levels(T)] * d))

    # Stack all nd-grid coordinates into one long matrix, of shape (T**d, d)
    U: Array = np.stack([U_grid.reshape(-1) for U_grid in U_grids], axis=1)
    assert U.shape == (Td, d)

    return U


def decode_quantile_values(T: int, d: int, Y_hat: Array) -> Sequence[Array]:
    """
    Decodes the regression coefficients of a VQR solution into vector quantile values.
    :param T: The number of quantile levels that was used for solving the problem.
    :param d: The dimension of the target data (Y) that was used for solving the
        problem.
    :param Y_hat: The regression output, of shape (T**d, 1).
    :return: A sequence of length d of vector quantile values. Each element j in the
        sequence is a d-dimensional array of shape (T, T, ..., T) containing the
        vector quantiles values of j-th variable in Y, i.e., the quantiles of Y_j|Y_{-j}
        where Y_{-j} means all the variables in Y except the j-th.
    """
    Q = np.reshape(Y_hat, newshape=(T,) * d)

    Q_functions: List[Array] = [np.array([np.nan])] * d
    for axis in reversed(range(d)):
        # Calculate derivative along this axis
        dQ_du = (1 / T) * np.diff(Q, axis=axis)

        # Duplicate first "row" along axis and insert it first
        pad_with = [
            (0, 0),
        ] * d
        pad_with[axis] = (1, 0)
        dQ_du = np.pad(dQ_du, pad_width=pad_with, mode="edge")

        Q_functions[d - 1 - axis] = dQ_du * T**2

    return tuple(Q_functions)


def decode_quantile_grid(T: int, d: int, U: Array) -> Sequence[Array]:
    """
    Decodes stacked vector quantile levels (the output of the vector_quantile_level()
    function) into a "meshgrid" of the evaluation points of the vector quantile
    function.

    :param T: The number of quantile levels that was used for solving the problem.
    :param d: The dimension of the target data (Y) that was used for solving the
    problem.
    :param U: The encoded grid U, of shape (T**d, d).
    :return: A sequence length d. Each array in the sequence is a d-dimensional
    array of shape (T, T, ..., T), which together represent the d-dimensional grid
    on which the vector quantiles were evaluated.
    """
    return tuple(np.reshape(U[:, dim], newshape=(T,) * d) for dim in range(d))


def inversion_sampling(T: int, d: int, n: int, Qs: Sequence[Array]):
    """
    Generates samples from the variable Y based on it's fitted
    quantile function, using inversion-transform sampling.
    :param T: The number of quantile levels that was used for solving the problem.
    :param d: The dimension of the target data (Y) that was used for solving the
        problem.
    :param n: Number of samples to generate.
    :param Qs: Quantile surfaces per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :return: Samples obtained from this quantile function, of shape (n, d).
    """

    # Samples of points on the quantile-level grid
    Us = np.random.randint(0, T, size=(n, d))

    # Sample from Y|X=x
    Y_samp = np.empty(shape=(n, d))
    for i, U_i in enumerate(Us):
        # U_i is a vector-quantile level, of shape (d,)
        Y_samp[i, :] = np.array([Q_d[tuple(U_i)] for Q_d in Qs])

    return Y_samp


def quantile_contour(
    T: int, d: int, Qs: Sequence[Array], alpha: float = 0.05
) -> Tuple[Array, Array]:
    """
    Creates a contour of points in d-dimensional space which surround the region in
    which 1-2*alpha of the distribution (that was corresponds to a given quantile
    function) is contained.

    :param T: The number of quantile levels that was used for solving the problem.
    :param d: The dimension of the target data (Y) that was used for solving the
        problem.
    :param Qs: Quantile surfaces per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :param alpha: Confidence level for the contour.
    :return: A tuple of two arrays, each of shape (n, d).
    The first contains the points along the d-dimensional
    contour, and the second contains the d-dimensional index of each point.
    """

    if not 0 < alpha < 0.5:
        raise ValueError(f"Got {alpha=}, but must be in (0, 0.5)")

    lo = int(np.round(T * alpha))
    hi = int(min(np.round(T * (1 - alpha)), T - 1))

    # Will contain d lists of points, each element will be (d,)
    contour_points_list = [[] for _ in range(d)]
    contour_index_list = [[] for _ in range(d)]

    for i, Q in enumerate(Qs):
        for j in range(d):
            for lo_hi, _ in zip([lo, hi], range(d)):
                # The zip here is just to prevent a loop when d==1.
                # This keeps all dims fixed on either lo or hi, while the j-th
                # dimension is sweeped from lo to hi.
                # idx: List[Union[int, slice]] = [lo_hi] * d
                # idx[j] = slice(lo, hi)

                idxs = [repeat(lo_hi)] * d
                idxs[j] = range(lo, hi)
                zipped_idxs = tuple(zip(*idxs))  # list of tuples e.g. [(u1, u2), ...]
                # d lists, e.g. [(u1, u1, ...), (u2, u2, ...)]
                grouped_indices = tuple(zip(*zipped_idxs))

                contour_points_list[i].extend(Q[grouped_indices])
                contour_index_list[i].extend(zipped_idxs)

    # Contours generated by accessing the same indices in each quantile surface
    assert all(
        np.allclose(contour_index_list[0], contour_index_list[i]) for i in range(1, d)
    )
    contour_points = np.array(contour_points_list).T  # (N, d)
    contour_indices = np.array(contour_index_list[0])  # (N, d)
    return contour_points, contour_indices


def vector_monotone_rearrangement(
    T: int, d: int, Qs: Sequence[Array], max_iters: int = 2e6
) -> Sequence[Array]:
    """
    Performs vector monotone rearrangement. Can be interpreted as "vector sorting".

    A vector-extension for the quantile rearrangement idea proposed in

    Victor Chernozhukov, Iván Fernández‐Val, Alfred Galichon.
    Quantile and probability curves without crossing.
    Econometrica, 2010.

    Solves an exact OT problem using POT's emd solver with U @ Y.T as the cost matrix.
    Rearrangement is performed by multiplication with the permutation matrix Pi.

    :param T: The number of quantile levels that was used for solving the problem.
    :param d: The dimension of the target data (Y) that was used for solving the
        problem.
    :param Qs: Quantile surfaces per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :param max_iters: Maximum number of iterations for the emd solver.
    :return: Rearranged quantile surfaces per dimension of Y. A sequence of length d
    where each element is of shape (T, T, ..., T).
    """
    U: Array = vector_quantile_levels(T, d)
    Y: Array = np.stack([Q.ravel() for Q in Qs], axis=-1)
    pi = ot.emd(M=-U @ Y.T, a=[], b=[], numItermax=max_iters)
    rearranged_Y = (T**d) * pi @ Y
    rearranged_Qs = [rearranged_Y[:, i].reshape((T,) * d) for i in range(d)]
    return rearranged_Qs


def check_comonotonicity(
    T: int, d: int, Qs: Sequence[Array], Us: Sequence[Array]
) -> Array:
    """
    Measures co-monotonicity defined as (u_i - u_j).T @ (Q(u_i) - Q(u_j)) for all i, j.
    Results in a T^d x T^d symmetric matrix.  If co-monotonicity is satisfied,
    all entries in this matrix are positive. Negative entries in this matrix
    represent the quantile crossing problem (and its analogue in higher dimensions).

    :param T: The number of quantile levels that was used for solving the problem.
    :param d: The dimension of the target data (Y) that was used for solving the
        problem.
    :param Qs: Quantile surfaces per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :param Us: Quantile levels per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :return: A T^d x T^d symmetric matrix that measures co-monotonicity between all
    pairs of quantile levels and quantile values. The (i, j)^th entry in the matrix
    measures co-monotonicity between quantile levels  u_i and u_j and the quantile
    values Q(u_i) and Q(u_j).
    """
    levels = np.stack([level.ravel() for level in Us])
    quantiles = np.stack([Q.ravel() for Q in Qs])

    pairwise_diff_levels = (
        levels.reshape(d, 1, T**d) - levels.reshape(d, T**d, 1)
    ).reshape(d, T ** (2 * d))

    pairwise_diff_quantiles = (
        quantiles.reshape(d, 1, T**d) - quantiles.reshape(d, T**d, 1)
    ).reshape(d, T ** (2 * d))

    all_pairs_inner_prod = np.sum(
        pairwise_diff_levels * pairwise_diff_quantiles, axis=0
    ).reshape(T**d, T**d)

    return all_pairs_inner_prod



import logging
from typing import Any, Dict

import click
import numpy as np
from matplotlib import pyplot as plt

from vqr import QuantileFunction, VectorQuantileRegressor
from experiments.base import VQROptions, run_exp_context
from experiments.datasets.mvn import LinearMVNDataProvider
from experiments.datasets.quantile import QuantileFunctionDataProviderWrapper

_LOG = logging.getLogger(__name__)


def _compare_conditional_quantiles(
    vqf_gt: QuantileFunction,
    vqf_est: QuantileFunction,
    t_factor: int,
) -> float:

    # Make sure shapes are consistent and that both quantile functions are
    # conditional on the same X.
    assert vqf_gt.d == vqf_est.d
    assert t_factor == vqf_gt.T // vqf_est.T
    assert np.allclose(vqf_gt.X, vqf_est.X)

    T = vqf_est.T
    d = vqf_est.d

    # We support having more levels in the gt function, in which case we index the
    # levels corresponding to the estimated levels.
    idx = (slice(None), *[slice(t_factor - 1, None, t_factor)] * d)

    q_surfaces_gt = vqf_gt.values[idx]  # (d, T, T, ..., T)
    q_surfaces_gt = q_surfaces_gt.reshape(d, -1)
    q_surfaces_est = vqf_est.values  # (d, T, T, ..., T)
    q_surfaces_est = q_surfaces_est.reshape(d, -1)
    assert q_surfaces_gt.shape == q_surfaces_est.shape

    # Make sure we're comparing the same quantile levels
    q_levels_gt = vqf_gt.levels[idx].reshape(d, -1)
    q_levels_est = vqf_est.levels.reshape(d, -1)
    assert q_levels_gt.shape == q_levels_est.shape
    assert np.allclose(q_levels_gt, q_levels_est)

    return (
        np.linalg.norm(q_surfaces_gt - q_surfaces_est) / np.linalg.norm(q_surfaces_gt)
    ).item()


def single_optim_exp(
    N: int,
    T: int,
    d: int,
    k: int,
    solver_name: str,
    solver_opts: dict,
    dp_vqr_n: int,
    dp_vqr_t_factor: int,
    dp_vqr_solver_opts: dict,
    n_eval_x: int,
    seed: int = 42,
) -> Dict[str, Any]:

    dp_vqr_solver_opts["verbose"] = True
    # Data provider
    wrapped_provider = LinearMVNDataProvider(d=d, k=k, seed=seed)
    data_provider = QuantileFunctionDataProviderWrapper(
        wrapped_provider=wrapped_provider,
        vqr_n_levels=T * dp_vqr_t_factor,
        vqr_fit_n=dp_vqr_n,
        vqr_solver_opts=dp_vqr_solver_opts,
        seed=seed,
    )

    # Sample values of x on which we evaluate
    eval_x = wrapped_provider.sample_x(n=n_eval_x)

    # Obtain g.t. VQR quantile functions
    vqfs_gt = data_provider.vqr.vector_quantiles(X=eval_x, refine=True)

    # Generate data
    X, Y = data_provider.sample(n=N)

    optimization_dists = []

    # Define a callback that will be invoked each iteration (epoch or batch)
    def _post_iter_callback(
        solution,
        batch_loss,
        epoch_loss,
        epoch_idx,
        batch_idx,
        num_epochs,
        num_batches,
        xy_slice,
        u_slice,
    ):
        # Obtain quantile functions from current iteration, conditioned on the same X's
        eval_x_scaled = data_provider.vqr._scaler.transform(eval_x)
        vqfs_est = solution.vector_quantiles(X=eval_x_scaled, refine=True)

        # Calculate distance from g.t.
        dists = [
            _compare_conditional_quantiles(vqf_gt, vqf_est, t_factor=dp_vqr_t_factor)
            for vqf_gt, vqf_est in zip(vqfs_gt, vqfs_est)
        ]
        optimization_dists.append(dists)

    # Add a solver callback to evaluate the distance from g.t. solution
    solver_opts["post_iter_callback"] = _post_iter_callback

    ###
    solver_opts["verbose"] = True

    # Solve
    vqr = VectorQuantileRegressor(
        n_levels=T,
        solver=solver_name,
        solver_opts=solver_opts,
    ).fit(X, Y)

    # Remove callback so that it doesn't get serialized
    solver_opts.pop("post_iter_callback")

    if plot := False:
        x = np.arange(len(optimization_dists))
        y = np.mean(optimization_dists, axis=1)
        yerr_min = np.min(optimization_dists, axis=1)
        yerr_max = np.max(optimization_dists, axis=1)

        fig, ax = plt.subplots(1, 1, figsize=(10, 5))
        ax.plot(x, y, "k-")
        ax.grid()
        ax.set_ylim([0, 2])
        ax.fill_between(x, y - yerr_min, y + yerr_max, alpha=0.5)
        ax.set_xlabel("epoch")
        ax.set_ylabel(r"$\frac{||Q^{*}(u)-\hat{Q}(u)||_{2}}{||Q^{*}(u)||_{2}}$")
        plt.savefig("optim-exp.png", tight_layout=True)

    return dict(
        N=N,
        T=T,
        d=d,
        k=k,
        solver_type=solver_name,
        solver=solver_opts,  # note: non-consistent key name on purpose
        optimization_dists=np.round(optimization_dists, decimals=5).tolist(),
    )


@click.command(name="optim-exp")
@click.pass_context
@VQROptions.cli
@VQROptions.cli(prefix="dp")
@click.option(
    "--t-factor",
    type=int,
    default=100,
    help=(
        "Factor between data provider's number of quantile levels and the fitted "
        "models' number of levels. Data provider will have T*t_factor levels."
    ),
)
@click.option(
    "--n-eval-x",
    type=int,
    default=10,
    help=("Number of X's to use when comparing g.t. to estimated quantile function"),
)
def optim_exp(
    ctx: click.Context,
    t_factor: int,
    n_eval_x: int,
    **kw,
):

    # We don't support multiple DP options, only first is used
    vqr_options_dp = VQROptions.parse(ctx, prefix="dp")

    # Generate experiment configs from CLI
    vqr_options = VQROptions.parse_multiple(ctx)
    exp_configs = {
        vqr_option.key(): dict(
            n_eval_x=n_eval_x,
            dp_vqr_n=vqr_options_dp.N,
            dp_vqr_t_factor=t_factor,
            dp_vqr_solver_opts=vqr_options_dp.solver_opts,
            **vqr_option.to_dict(),
        )
        for vqr_option in vqr_options
    }

    return run_exp_context(
        ctx, exp_fn=single_optim_exp, exp_configs=exp_configs, write_csv=True
    )



# isort: skip_file
from vqr.vqr import VQRSolver, VQRSolution, QuantileFunction
from vqr.api import VectorQuantileEstimator, VectorQuantileRegressor
from pkg_resources import DistributionNotFound, get_distribution

try:
    dist_name = "vqr"
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound



import os
import pickle
import logging
from typing import Tuple, Optional
from pathlib import Path

import numpy as np
from filelock import FileLock

from vqr.api import DEFAULT_SOLVER_NAME, VectorQuantileRegressor
from experiments import EXPERIMENTS_OUT_DIR
from experiments.datasets.base import Array, DataProvider
from experiments.utils.helpers import stable_hash

_LOG = logging.getLogger(__name__)

DEFAULT_CACHE_DIR = EXPERIMENTS_OUT_DIR.joinpath("cache")


class QuantileFunctionDataProviderWrapper(DataProvider):
    """
    A wrapper around a data provider which first estimates the conditional quantile
    function by solving VQR, and then samples from the quantile function to generate
    new data.

    Uses a local cache to avoid re-calculating the same VQR solution multiple times.
    """

    def __init__(
        self,
        wrapped_provider: DataProvider,
        vqr_n_levels: int = 50,
        vqr_fit_n: int = 10000,
        vqr_solver_name: str = DEFAULT_SOLVER_NAME,
        vqr_solver_opts: Optional[dict] = None,
        cache_dir: Optional[Path] = DEFAULT_CACHE_DIR,
        seed: Optional[int] = 42,
    ):
        """
        :param wrapped_provider: The data provider which will be used to draw samples
        on which a VQR model will be fitted.
        :param vqr_n_levels: Number of quantile levels to fit the VQR model with.
        :param vqr_fit_n: Number of samples from the wrapped provider to fit the VQR
        model with.
        :param vqr_solver_name: VQR solver name.
        :param vqr_solver_opts: VQR solver options.
        :param cache_dir: Override for the default cache directory. Set None for no
        caching.
        :param seed: Random seed.
        """
        super().__init__(seed=seed)

        self.wrapped_provider = wrapped_provider
        self._vqr_n_levels = vqr_n_levels
        self._vqr_fit_n = vqr_fit_n
        self._vqr_solver_name = vqr_solver_name
        self._vqr_solver_opts = vqr_solver_opts
        self._cache_dir = Path(cache_dir) if cache_dir else None

        if not self._cache_dir:
            self.vqr = self._fit_vqr()

        else:
            (
                self.vqr,
                self.cached_path,
                self.cache_loaded,
                self.cache_saved,
            ) = self._vqr_from_cache()

    def _fit_vqr(self) -> VectorQuantileRegressor:
        X, Y = self.wrapped_provider.sample(n=self._vqr_fit_n)
        return VectorQuantileRegressor(
            n_levels=self._vqr_n_levels,
            solver=self._vqr_solver_name,
            solver_opts=self._vqr_solver_opts,
        ).fit(X, Y)

    def _vqr_from_cache(self):
        os.makedirs(self._cache_dir, exist_ok=True)

        cache_args = [
            self.wrapped_provider,
            self._vqr_fit_n,
            self._vqr_n_levels,
            self._vqr_solver_name,
            self._vqr_solver_opts,
            self._seed,
        ]
        cache_key = stable_hash(cache_args, hash_len=8)

        cached_file_path = self._cache_dir.joinpath(f"{cache_key}.pkl")
        lock_file_path = self._cache_dir.joinpath(f"{cache_key}.lock")

        # Make sure only one process calculates the same cached VQR
        with FileLock(str(lock_file_path), timeout=-1):
            need_to_refit = True
            cache_loaded = False
            cache_saved = False

            # First we try to load from cache
            if cached_file_path.is_file():
                with open(cached_file_path, "rb") as f:
                    try:
                        fitted_vqr = pickle.load(f)
                        _LOG.info(f"Loaded cached VQR@{cache_key}")
                        need_to_refit = False
                        cache_loaded = True
                    except pickle.PickleError as e:
                        _LOG.warning(f"Failed to load cached VQR@{cache_key}: {e}")

            # If there is no cached result, or we were unsuccessful loading it,
            # fit VQR and save the cached result.
            if need_to_refit:
                fitted_vqr = self._fit_vqr()

                with open(cached_file_path, "wb") as f:
                    pickle.dump(fitted_vqr, f)
                    _LOG.info(f"Saved cached VQR@{cache_key}")
                    cache_saved = True

        return fitted_vqr, cached_file_path, cache_loaded, cache_saved

    @property
    def k(self) -> int:
        return self.wrapped_provider.k

    @property
    def d(self) -> int:
        return self.wrapped_provider.d

    def sample_x(self, n: int) -> Array:
        return self.wrapped_provider.sample_x(n)

    def sample(self, n: int, x: Optional[Array] = None) -> Tuple[Array, Array]:
        if x is None:
            X = self.sample_x(n=n)
            Y = np.concatenate([self.vqr.sample(n=1, x=x) for x in X], axis=0)
        else:
            x = np.reshape(x, (1, -1))
            assert x.shape[1] == self.k
            X = np.concatenate([x for _ in range(n)], axis=0)
            Y = self.vqr.sample(n=n, x=x)

        return X, Y



import numpy as np
from sklearn.datasets import make_regression
from sklearn.model_selection import train_test_split

from vqr import VectorQuantileRegressor
from vqr.solvers.dual.regularized_lse import RegularizedDualVQRSolver


def test_vqr_minimal():
    N, d, k, T = 5000, 2, 1, 20
    N_test = N // 10
    seed = 42
    alpha = 0.05

    # Generate some data (or load from elsewhere).
    X, Y = make_regression(
        n_samples=N, n_features=k, n_targets=d, noise=0.1, random_state=seed
    )
    X_train, X_test, Y_train, Y_test = train_test_split(
        X, Y, test_size=N_test, shuffle=True, random_state=seed
    )

    # Create the VQR solver and regressor.
    vqr_solver = RegularizedDualVQRSolver(
        verbose=True, epsilon=1e-2, num_epochs=1000, lr=0.9
    )
    vqr = VectorQuantileRegressor(n_levels=T, solver=vqr_solver)

    # Fit the model on the data.
    vqr.fit(X_train, Y_train)

    # Conditional coverage calculation: for each test point, calculate the
    # conditional quantiles given x, and check whether the corresponding y is covered
    # in the alpha-contour.
    cov_test = np.mean(
        [vqr.coverage(Y_test[[i]], X_test[[i]], alpha=alpha) for i in range(N_test)]
    )
    print(f"{cov_test=}")

    # Sample from the fitted conditional distribution, given a specific x.
    Y_sampled = vqr.sample(n=100, x=X_test[0])

    # Calculate coverage on the samples.
    cov_sampled = vqr.coverage(Y_sampled, x=X_test[0], alpha=alpha)
    print(f"{cov_sampled=}")



from __future__ import annotations

import logging
import functools
from abc import ABC, abstractmethod
from time import strftime
from typing import Any, Dict, List, Callable, Optional, Sequence
from pathlib import Path
from itertools import product

import click
import pandas as pd
from _socket import gethostname

from experiments import EXPERIMENTS_OUT_DIR
from experiments.utils.helpers import stable_hash
from experiments.utils.parallel import run_parallel_exp
from vqr.solvers.dual.regularized_lse import (
    RegularizedDualVQRSolver,
    MLPRegularizedDualVQRSolver,
)

_LOG = logging.getLogger(__name__)


def _click_decorator(fn: Callable, options: Sequence[click.Option]):
    """
    Generates a decorator which applies a sequence of click.Options to function.
    :param fn: The function to decorate.
    :param options: The click options to apply to the function.
    :return: The decorated function.
    """
    # Reduce applies each option to f repeatedly, i.e. opt1(opt2(...optN(f) ...))
    wrapper = functools.reduce(lambda fn_, opt: opt(fn_), reversed(options), fn)
    return functools.wraps(fn)(wrapper)


class _CLIOptions(ABC):
    """
    Helper class for grouping multiple click CLI options together and then parsing them
    from the context.
    """

    @classmethod
    @abstractmethod
    def _cli_options(cls, prefix: str = "") -> Sequence[click.Option]:
        """
        :param prefix: A prefix to add to each argument name.
        :return: A list of click.Option decorators which can be applied to a
        click.Command.
        """
        pass

    @classmethod
    def cli(cls, fn: Callable = None, *, prefix: str = None) -> Callable:
        """
        A decorator which can be applied to a click.Command. Adds the options defined in
        the _cli_options method.
        :param fn: The function to decorate, should be a click command or group.
        :param prefix: A prefix which will be applied to all CLI options.
        :return: A decorator which adds click options to the given command/group.
        """

        prefix = prefix or ""

        # This is to support the usage as a parametrized decorator, i.e. @cli(k=v)
        if fn is None:
            return functools.partial(cls.cli, prefix=prefix)

        return _click_decorator(fn, cls._cli_options(prefix=prefix))

    @classmethod
    def parse(cls, ctx: click.Context, prefix: str = "") -> _CLIOptions:
        """
        Parses a click.Context into a _CLIOptions instance.
        If there are multiple option values, returns the first one.
        :param ctx: The click.Context.
        :param prefix: The prefix which was applied to all CLI options.
        :return: An instance of this class, initialized based on parsing the context.
        """
        multi_opts = cls.parse_multiple(ctx, prefix=prefix)
        n_opts = len(multi_opts)
        if n_opts != 1:
            raise ValueError(f"Expected single value of parsed options, got {n_opts}")

        return multi_opts[0]

    @classmethod
    @abstractmethod
    def parse_multiple(
        cls,
        ctx: click.Context,
        prefix: str = "",
    ) -> Sequence[_CLIOptions]:
        """
        Parses a click.Context, which may contain some options with multiple
        values, into a sequence of _CLIOptions instance, generated as a product of
        the multiple values.
        :param ctx: The click.Context.
        :param prefix: The prefix which was applied to all CLI options.
        :return: An sequence of instances of this class, initialized based on parsing
        the context.
        """
        pass

    @classmethod
    def _prefix_option(cls, option: str, prefix: str):
        if not prefix:
            return option

        if option.startswith("--"):
            return option.replace("--", f"--{prefix}-")
        elif option.startswith("-"):
            return option.replace("-", f"-{prefix}_")
        else:
            raise ValueError(
                f"Expected that options start with '-' or '--', got {option}"
            )

    @classmethod
    def _prefix_name(cls, name: str, prefix: str):
        if not prefix:
            return name

        if name.startswith("-"):
            raise ValueError(
                f"Expected that option names don't start with '-', got {name}"
            )

        return f"{prefix}_{name}"

    def key(self) -> str:
        """
        :return: A string which uniquely represents the current instance's options.
        """
        return stable_hash(self.__dict__)

    def __hash___(self):
        return hash(self.key())


class GPUOptions(_CLIOptions):
    """
    Encapsulates GPU options for an experiment.
    """

    def __init__(
        self,
        gpu_enabled: bool,
        gpu_devices: Optional[str],
        num_processes: int,
        ppd: int,
    ):
        self.gpu_enabled = gpu_enabled
        self.gpu_devices = gpu_devices
        self.num_processes = num_processes
        self.ppd = ppd

    @classmethod
    def _cli_options(cls, prefix: str = "") -> Sequence[click.Option]:
        _p = functools.partial(cls._prefix_option, prefix=prefix)
        return [  # type:ignore
            click.option(
                _p("-g"), _p("--gpu/--no-gpu"), default=False, help="Enable GPU support"
            ),
            click.option(
                _p("--devices"),
                default=None,
                type=str,
                help="GPU devices (comma separated)",
            ),
            click.option(
                _p("-p"),
                _p("--processes"),
                type=int,
                default=-1,
                help="Number of processes",
            ),
            click.option(
                _p("--ppd"), type=int, default=1, help="Processes per GPU device"
            ),
        ]

    @classmethod
    def parse_multiple(
        cls,
        ctx: click.Context,
        prefix: str = "",
    ) -> Sequence[GPUOptions]:
        _n = functools.partial(cls._prefix_name, prefix=prefix)

        while ctx and _n("gpu") not in ctx.params:
            ctx = ctx.parent
        if not ctx:
            raise ValueError(f"GPU options not found in context ({prefix=})")

        return (
            GPUOptions(
                gpu_enabled=ctx.params[_n("gpu")],
                gpu_devices=ctx.params[_n("devices")],
                num_processes=ctx.params[_n("processes")],
                ppd=ctx.params[_n("ppd")],
            ),
        )


class OutputOptions(_CLIOptions):
    """
    Encapsulates output options for an experiment.
    """

    def __init__(self, out_dir: Path, out_tag: str):
        self.out_dir = out_dir
        self.out_tag = out_tag

    @classmethod
    def _cli_options(cls, prefix: str = "") -> Sequence[click.Option]:
        _p = functools.partial(cls._prefix_option, prefix=prefix)
        return [  # type:ignore
            click.option(
                _p("-o"),
                _p("--out-dir"),
                type=Path,
                default=EXPERIMENTS_OUT_DIR,
                help="Output directory",
            ),
            click.option(_p("--out-tag"), type=str, default="", help="Output tag"),
        ]

    @classmethod
    def parse_multiple(
        cls,
        ctx: click.Context,
        prefix: str = "",
    ) -> Sequence[OutputOptions]:
        _n = functools.partial(cls._prefix_name, prefix=prefix)

        while ctx and _n("out_dir") not in ctx.params:
            ctx = ctx.parent
        if not ctx:
            raise ValueError(f"Output options not found in context {prefix=}")

        return (
            OutputOptions(
                out_dir=ctx.params[_n("out_dir")],
                out_tag=ctx.params[_n("out_tag")],
            ),
        )


class LoggingOptions(_CLIOptions):
    """
    Encapsulates output options for an experiment.
    """

    def __init__(self, debug: bool):
        self.debug = debug

    @classmethod
    def _cli_options(cls, prefix: str = "") -> Sequence[click.Option]:
        _p = functools.partial(cls._prefix_option, prefix=prefix)
        return [  # type:ignore
            click.option(
                _p("--debug/--no-debug"), default=False, help="Use DEBUG logging"
            ),
        ]

    @classmethod
    def parse_multiple(
        cls,
        ctx: click.Context,
        prefix: str = "",
    ) -> Sequence[LoggingOptions]:
        _n = functools.partial(cls._prefix_name, prefix=prefix)

        while ctx and _n("debug") not in ctx.params:
            ctx = ctx.parent
        if not ctx:
            raise ValueError(f"Logging options not found in context {prefix=}")

        return (LoggingOptions(debug=ctx.params[_n("debug")]),)


class VQROptions(_CLIOptions):
    """
    Encapsulates VQR options for an experiment.
    """

    def __init__(
        self,
        N: int,
        T: int,
        d: int,
        k: int,
        solver_name: str,
        solver_opts: Dict[str, Any],
    ):
        self.N = N
        self.T = T
        self.d = d
        self.k = k
        self.solver_name = solver_name
        self.solver_opts = solver_opts

    def to_dict(self):
        return self.__dict__.copy()

    @classmethod
    def _cli_options(cls, prefix: str = "") -> Sequence[click.Option]:
        _p = functools.partial(cls._prefix_option, prefix=prefix)
        _n = functools.partial(cls._prefix_name, prefix=prefix)

        return [  # type:ignore
            click.option(
                _p("-N"),
                _n("ns"),
                type=int,
                multiple=True,
                default=[1000],
                help="Samples",
            ),
            click.option(
                _p("-T"),
                _n("ts"),
                type=int,
                multiple=True,
                default=[20],
                help="Quantile levels",
            ),
            click.option(
                _p("-d"),
                _n("ds"),
                type=int,
                multiple=True,
                default=[2],
                help="Y dimension",
            ),
            click.option(
                _p("-k"),
                _n("ks"),
                type=int,
                multiple=True,
                default=[3],
                help="X dimension",
            ),
            click.option(
                _p("-E"),
                _n("epsilons"),
                type=float,
                multiple=True,
                default=[1e-6],
                help="epsilon",
            ),
            click.option(
                _p("--bs-y"),
                _n("bys"),
                type=int,
                multiple=True,
                default=[-1],
                help="Batch size of Y (samples), -1 disables batching",
            ),
            click.option(
                _p("--bs-u"),
                _n("bus"),
                type=int,
                multiple=True,
                default=[-1],
                help="Batch size of U (quantile levels), -1 disables batching",
            ),
            click.option(_p("--epochs"), type=int, default=1000, help="epochs"),
            click.option(_p("--lr"), type=float, default=0.5, help="Learning rate"),
            click.option(
                _p("--lr-max-steps"), type=int, default=10, help="LR sched. steps"
            ),
            click.option(
                _p("--lr-factor"), type=float, default=0.9, help="LR sched. factor"
            ),
            click.option(
                _p("--lr-patience"), type=int, default=500, help="LR sched. patience"
            ),
            click.option(
                _p("--lr-threshold"),
                type=float,
                default=0.01 * 5,
                help="LR sched. " "thresh.",
            ),
            click.option(
                _p("--mlp/--no-mlp"), type=bool, default=False, help="NL-VQR with MLP"
            ),
            click.option(
                _p("--mlp-layers"),
                type=str,
                default="32,32",
                help="comma-separated ints",
            ),
            click.option(
                _p("--mlp-skip/--no-mlp-skip"),
                type=bool,
                default=False,
                help="MLP residual",
            ),
            click.option(
                _p("--mlp-activation"), type=str, default="relu", help="MLP activation"
            ),
            click.option(
                _p("--mlp-batchnorm/--no-mlp-batchnorm"), type=bool, default=False
            ),
            click.option(_p("--mlp-dropout"), type=float, default=0.0),
        ]

    @classmethod
    def parse_multiple(
        cls,
        ctx: click.Context,
        prefix: str = "",
    ) -> Sequence[VQROptions]:
        _n = functools.partial(cls._prefix_name, prefix=prefix)

        while ctx and _n("ns") not in ctx.params:
            ctx = ctx.parent
        if not ctx:
            raise ValueError(f"VQR options not found in context {prefix=}")

        gpu_options = GPUOptions.parse(ctx)
        p = ctx.params

        # parse mlp options
        mlp_opts = dict(
            hidden_layers=p[_n("mlp_layers")],
            activation=p[_n("mlp_activation")],
            skip=p[_n("mlp_skip")],
            batchnorm=p[_n("mlp_batchnorm")],
            dropout=p[_n("mlp_dropout")],
        )
        # Filter out defaults and remove everything if mlp==False.
        mlp_opts = {k: v for k, v in mlp_opts.items() if v is not None and p[_n("mlp")]}

        solver_name = (
            MLPRegularizedDualVQRSolver.solver_name()
            if p[_n("mlp")]
            else RegularizedDualVQRSolver.solver_name()
        )

        vqr_options = tuple(
            VQROptions(
                N=N_,
                T=T_,
                d=d_,
                k=k_,
                solver_name=solver_name,
                solver_opts=dict(
                    verbose=False,
                    num_epochs=p[_n("epochs")],
                    epsilon=eps_,
                    lr=p[_n("lr")],
                    lr_max_steps=p[_n("lr_max_steps")],
                    lr_factor=p[_n("lr_factor")],
                    lr_patience=p[_n("lr_patience")],
                    lr_threshold=p[_n("lr_threshold")],
                    batchsize_y=bs_y_ if bs_y_ > 0 else None,
                    batchsize_u=bs_u_ if bs_u_ > 0 else None,
                    gpu=gpu_options.gpu_enabled,
                    **mlp_opts,
                ),
            )
            for (N_, T_, d_, k_, eps_, bs_y_, bs_u_) in product(
                p[_n("ns")],
                p[_n("ts")],
                p[_n("ds")],
                p[_n("ks")],
                p[_n("epsilons")],
                p[_n("bys")],
                p[_n("bus")],
            )
        )

        return vqr_options


def experiment_id(name: str, tag: str):
    """
    Creates a unique id for an experiment based on hostname, timestamp and a
    user-specified tag.
    :param name: An experiment name.
    :param tag: A user tag.
    :return: The experiment id.
    """
    hostname = gethostname()
    if hostname:
        hostname = hostname.split(".")[0].strip()
    else:
        hostname = "localhost"

    name = f"{name}-" if name else ""
    tag = f"-{tag}" if tag else ""
    timestamp = strftime(f"%Y%m%d_%H%M%S")
    exp_id = strftime(f"{name}{timestamp}-{hostname}{tag}")
    return exp_id


def run_exp_context(
    ctx: click.Context,
    exp_fn: Callable[[Any], Dict[str, Any]],
    exp_configs: Dict[str, dict],
    write_csv: bool = True,
) -> pd.DataFrame:
    """
    Runs multiple experiment configs based on common CLI arguments from the context.
    :param ctx: Click context.
    :param exp_fn: Callable which runs a single experiment. Should return a dict of
    results.
    :param exp_configs:  A mapping from an experiment name/identifier to a dict with
    configuration which will be passed to the exp_fn.
    :param write_csv: Whether to write the results to a CSV.
    :return: A pandas DataFrame with the collected results. Each row is a result,
    columns correspond to dict keys in the output of exp_fn.
    """
    output_opts = OutputOptions.parse(ctx)
    gpu_opts = GPUOptions.parse(ctx)
    exp_id = experiment_id(name=ctx.command.name, tag=output_opts.out_tag)

    results: Sequence[Dict[str, Any]] = run_parallel_exp(
        exp_name=exp_id,
        exp_fn=exp_fn,
        exp_configs=exp_configs,
        max_workers=gpu_opts.num_processes,
        gpu_enabled=gpu_opts.gpu_enabled,
        gpu_devices=gpu_opts.gpu_devices,
        workers_per_device=gpu_opts.ppd,
    )

    results_df = pd.json_normalize(list(results))

    if write_csv:
        out_file_path = output_opts.out_dir.joinpath(f"{exp_id}.csv")
        results_df.to_csv(out_file_path, index=False)
        _LOG.info(f"Wrote output file: {out_file_path.absolute()!s}")

    return results_df



from typing import Dict

from numpy import array, argsort
from numpy.linalg import pinv

from vqr import VectorQuantileEstimator
from vqr.api import ScalarQuantileEstimator
from vqr.coverage import measure_width, measure_coverage
from experiments.utils.split import split_train_calib_test
from experiments.datasets.mvn import LinearMVNDataProvider


def experiment(
    X_: array, Y_: array, alpha_low_: float, alpha_high_: float, n_levels: int
) -> Dict[str, float]:
    assert X_.shape[0] == Y_.shape[0]
    assert alpha_low_ < alpha_high_
    assert 0 <= alpha_low_ <= 1
    assert 0 <= alpha_high_ <= 1

    # split train-valid-test
    splits = split_train_calib_test(X_, Y_, split_ratios=(0.15, 0.65))

    # Fit a linear model on train set
    X_train, Y_train = splits["train"]
    A_hat = pinv(X_train.T @ X_train) @ X_train.T @ Y_train

    # Calculate the non-conformity scores on the calibration set
    X_calib, Y_calib = splits["calib"]
    scores_calib = Y_calib - (X_calib @ A_hat)

    # Estimate vector quantiles on the calibration set
    vq = VectorQuantileEstimator(n_levels)
    fitted_vq = vq.fit(scores_calib)
    VQ1, VQ2 = fitted_vq.vector_quantiles()
    T = fitted_vq.quantile_levels

    # Construct a quantile contour
    idx_low = argsort(abs(T - alpha_low_))[0]
    idx_high = argsort(abs(T - alpha_high_))[0]

    vq_contour = array(
        [
            [
                *VQ1[idx_low : idx_high + 1, idx_high + 1],
                *VQ1[idx_low, idx_low : idx_high + 1],
                *VQ1[idx_low : idx_high + 1, idx_low],
                *VQ1[idx_high + 1, idx_low : idx_high + 1],
            ],
            [
                *VQ2[idx_low : idx_high + 1, idx_high + 1],
                *VQ2[idx_low, idx_low : idx_high + 1],
                *VQ2[idx_low : idx_high + 1, idx_low],
                *VQ2[idx_high + 1, idx_low : idx_high + 1],
            ],
        ]
    ).T

    # Fit scalar quantiles
    sq = ScalarQuantileEstimator(n_levels)
    fitted_sq = sq.fit(scores_calib)
    SQ = fitted_sq.vector_quantiles()
    SQ1 = SQ[:, 0]
    SQ2 = SQ[:, 1]
    sq_contour = array(
        [
            [
                *SQ1[idx_low : idx_high + 1],
                *SQ1[idx_low : idx_high + 1],
                *[SQ1[idx_low]] * len(SQ2[idx_low : idx_high + 1]),
                *[SQ1[idx_high + 1]] * len(SQ2[idx_low : idx_high + 1]),
            ],
            [
                *[SQ2[idx_high + 1]] * len(SQ1[idx_low : idx_high + 1]),
                *[SQ2[idx_low]] * len(SQ1[idx_low : idx_high + 1]),
                *SQ2[idx_low : idx_high + 1],
                *SQ2[idx_low : idx_high + 1],
            ],
        ]
    ).T

    # test dataset
    X_test, Y_test = splits["test"]

    vq_coverages_ = []
    sq_coverages_ = []
    for i in range(X_test.shape[0]):
        # prediction set for the data point i
        prediction_set_vq = vq_contour + X_test[i, :] @ A_hat
        prediction_set_sq = sq_contour + X_test[i, :] @ A_hat

        # measure coverage (0/1) for point i
        vq_coverages_.append(measure_coverage(prediction_set_vq, Y_test[i, :][None, :]))
        sq_coverages_.append(measure_coverage(prediction_set_sq, Y_test[i, :][None, :]))

    vq_width_ = measure_width(vq_contour)
    sq_width_ = measure_width(sq_contour)
    vq_coverage_ = sum(vq_coverages_) / len(vq_coverages_)
    sq_coverage_ = sum(sq_coverages_) / len(sq_coverages_)

    results_ = {
        "vq_width": vq_width_,
        "sq_width": sq_width_,
        "vq_coverage": vq_coverage_,
        "sq_coverage": sq_coverage_,
        "realized_alpha_low": T[idx_low],
        "realized_alpha_high": T[idx_high],
        "deflated_alpha_low": T[idx_low],
        "deflated_alpha_high": T[idx_high],
    }

    return results_


if __name__ == "__main__":
    num_trials = 100
    n = 1500
    d = 2
    k = 4
    alpha_low = 0.05
    alpha_high = 0.95
    n_levels = 40
    X, Y = LinearMVNDataProvider(d=d, k=k, seed=42).sample(n=n)

    vq_coverages = []
    vq_widths = []

    sq_coverages = []
    sq_widths = []
    for j in range(num_trials):
        results = experiment(X, Y, alpha_low, alpha_high, n_levels)
        vq_coverages.append(results["vq_coverage"])
        vq_widths.append(results["vq_width"])

        sq_coverages.append(results["sq_coverage"])
        sq_widths.append(results["sq_width"])

        print(
            f"Trial {j}: \n"
            f"Coverage level: [{round(results['realized_alpha_low'], 3)}, "
            f"{round(results['realized_alpha_high'], 3)}] \n"
            f"Vector quantiles: coverage={round(results['vq_coverage'], 3)}%, "
            f"width={round(results['vq_width'], 3)} \n"
            f"Deflated coverage level: [{round(results['deflated_alpha_low'], 3)}, "
            f"{round(results['deflated_alpha_high'], 3)}] \n"
            f"Scalar quantiles: coverage={round(results['sq_coverage'], 3)}%, "
            f"width={round(results['sq_width'], 3)} \n"
        )

    marginal_vq_coverage = sum(vq_coverages) / len(vq_coverages)
    average_vq_width = sum(vq_widths) / len(vq_widths)

    marginal_sq_coverage = sum(sq_coverages) / len(sq_coverages)
    average_sq_width = sum(sq_widths) / len(sq_widths)

    print(
        f"\t\t     Experiment summary \t\t \n"
        f" \t\t ========================= \t\t \n"
        f"Vector quantiles: \n"
        f"Marginal coverage over {num_trials} trials: "
        f"{marginal_vq_coverage}, average width = {average_vq_width}.\n"
        f"Scalar quantiles: \n"
        f"Marginal coverage over {num_trials} trials: "
        f"{marginal_sq_coverage}, average width = {average_sq_width}."
    )
    print(vq_coverages)
    print(vq_widths)
    print(sq_coverages)
    print(sq_widths)



import numpy as np
from numpy.typing import ArrayLike as Array
from scipy.spatial import ConvexHull


def measure_coverage(quantile_contour: Array, data: Array) -> float:
    """
    Measures the proportion of given points that lie in side a given surface.
    The surface is approximated by its convex hull, all points are checked
    if they lie within the surface or not.

    :param quantile_contour: points representing a quantile contour, of shape (M, d)
    :param data: the points that need to be checked if they are inliers, of shape (N, d)
    :return: Proportion of points that lie within the given contour.
    """

    if np.ndim(quantile_contour) != 2 or np.ndim(data) != 2:
        raise ValueError("Both input arrays must be 2d")

    N, d = data.shape

    if d == 1:
        coverage_ = (data >= quantile_contour[0]) & (data <= quantile_contour[-1])
    else:
        cvx_hull = ConvexHull(quantile_contour)
        coverage_ = np.array([point_in_hull(d, cvx_hull) for d in data])  # type: ignore

    return sum(coverage_).item() / len(coverage_)


def measure_width(quantile_contour: Array) -> float:
    """
    Measures the volume of the convex hull of the given array of points.
    :param quantile_contour: Points constituting the boundary of the contour over
    which the convex hull needs to be constructed, of shape (M, d).
    :return: The area (in 2D) / volume (in 3D) of the convex hull of the given points.
    """
    if np.ndim(quantile_contour) != 2:
        raise ValueError("Input must be 2d")

    _, d = quantile_contour.shape

    if d == 1:
        return quantile_contour[-1] - quantile_contour[0]

    cvx_hull = ConvexHull(quantile_contour)
    volume: float = cvx_hull.volume  # type: ignore
    return volume


def point_in_hull(point: Array, hull: ConvexHull, tolerance=1e-12) -> bool:
    """
    Returns whether a point is contained inside a convex hull.
    :param point: A point in n-d.
    :param hull: A ConvexHull object.
    :param tolerance: Tolerance for comparison.
    :return: True if the point is contained.
    """
    return all(
        (np.dot(eq[:-1], point) + eq[-1] <= tolerance)
        for eq in hull.equations  # type:ignore
    )



import os

import pytest

from tests import TESTS_OUT_DIR
from experiments.datasets.mvn import LinearMVNDataProvider, IndependentDataProvider
from experiments.datasets.quantile import QuantileFunctionDataProviderWrapper
from experiments.datasets.cond_banana import ConditionalBananaDataProvider


class TestMVNData(object):
    @pytest.mark.parametrize("k", [1, 20, 100])
    @pytest.mark.parametrize("d", [1, 2, 10])
    @pytest.mark.parametrize("n", [1000, 5000, 10000])
    def test_shapes(self, n, d, k):
        X, Y = IndependentDataProvider(k, d, seed=1023).sample(n=n)

        assert X.shape == (n, k)
        assert Y.shape == (n, d)


class TestCondBanana(object):
    @pytest.mark.parametrize("k", [1, 20, 100])
    @pytest.mark.parametrize("d", [2, 3, 4])
    @pytest.mark.parametrize("n", [1000, 5000, 10000])
    @pytest.mark.parametrize("nonlinear", [True, False])
    def test_shapes(self, n, d, k, nonlinear):
        X, Y = ConditionalBananaDataProvider(
            k=k, d=d, nonlinear=nonlinear, seed=63
        ).sample(n)

        assert X.shape == (n, k)
        assert Y.shape == (n, d)


class TestQuantileFunctionDataProviderWrapper:
    @pytest.fixture(autouse=True)
    def setup(self, test_out_dir):
        self.wrapped_provider = LinearMVNDataProvider(k=3, d=1)
        self.cache_dir = test_out_dir

    @pytest.fixture(autouse=True, scope="class")
    def providers(self, test_out_dir_class_pid):
        wrapped_provider = LinearMVNDataProvider(k=3, d=1, seed=5040)
        cache_dir = test_out_dir_class_pid

        vqr_n_levels = 25
        vqr_fit_n = 10000

        dp1 = QuantileFunctionDataProviderWrapper(
            wrapped_provider=wrapped_provider,
            vqr_solver_opts=dict(verbose=True),
            vqr_n_levels=vqr_n_levels,
            vqr_fit_n=vqr_fit_n,
            cache_dir=cache_dir,
        )

        dp1_cached = QuantileFunctionDataProviderWrapper(
            wrapped_provider=wrapped_provider,
            vqr_solver_opts=dict(verbose=True),
            vqr_n_levels=vqr_n_levels,
            vqr_fit_n=vqr_fit_n,
            cache_dir=cache_dir,
        )

        dp2 = QuantileFunctionDataProviderWrapper(
            wrapped_provider=wrapped_provider,
            vqr_solver_opts=dict(verbose=True),
            vqr_n_levels=vqr_n_levels // 2,
            vqr_fit_n=vqr_fit_n // 2,
            cache_dir=cache_dir,
        )

        return wrapped_provider, dp1, dp1_cached, dp2

    def test_cache(self, providers):
        wrapped_provider, dp1, dp1_cached, dp2 = providers

        assert not dp1.cache_loaded
        assert dp1.cache_saved
        assert dp1_cached.cache_loaded
        assert not dp1_cached.cache_saved

        assert not dp2.cache_loaded
        assert dp2.cache_saved

        assert dp1.cached_path == dp1_cached.cached_path
        assert dp1.cached_path != dp2

    def test_wrapped_shapes(self, providers):
        wrapped_provider, dp1, dp1_cached, dp2 = providers

        for dp in [dp1, dp1_cached, dp2]:
            assert dp.k == wrapped_provider.k
            assert dp.d == wrapped_provider.d

            n = 100

            for x in [None, dp.sample_x(n=1)]:
                if x is not None:
                    assert x.shape == (1, dp.k)

                X, Y = dp1.sample(n=n, x=x)
                assert Y.shape == (n, dp.d)
                assert X.shape == (n, dp.k)



from abc import ABC
from typing import Any, Dict, Type, Union, Optional, Sequence

import numpy as np
from numpy import ndarray as Array
from numpy import quantile
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils import check_X_y
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_is_fitted

from vqr.vqr import (
    VQRSolver,
    VQRSolution,
    QuantileFunction,
    quantile_levels,
    quantile_contour,
    inversion_sampling,
)
from vqr.coverage import measure_coverage
from vqr.solvers.primal.cvx import CVXVQRSolver
from vqr.solvers.primal.pot import POTVQESolver
from vqr.solvers.dual.regularized_lse import (
    RegularizedDualVQRSolver,
    MLPRegularizedDualVQRSolver,
)

SOLVER_TYPES: Dict[str, Type[VQRSolver]] = {
    solver_class.solver_name(): solver_class
    for solver_class in [
        CVXVQRSolver,
        RegularizedDualVQRSolver,
        MLPRegularizedDualVQRSolver,
        POTVQESolver,
    ]
}

DEFAULT_SOLVER_NAME = RegularizedDualVQRSolver.solver_name()


class VectorQuantileBase(BaseEstimator, ABC):
    """
    Base class for vector quantile estimation (VQE) and regression (VQR).
    Compatible with the sklearn API.
    """

    def __init__(
        self,
        n_levels: int = 50,
        solver: Union[str, VQRSolver] = DEFAULT_SOLVER_NAME,
        solver_opts: Optional[Dict[str, Any]] = None,
    ):
        """
        :param n_levels: Number of quantile levels. The range of possible
            levels (between 0 and 1) will be divided into this number of levels.
            For example, if n_levels=4 then the quantiles [.25, .50, .75, 1.0] will be
            evaluation in each dimension.
        :param solver: Either a supported solver name (see keys of
            :obj:`SOLVER_TYPES`) or an instance of a :class:`VQRSolver`.
        :param solver_opts: If solver is a string, these kwargs will be passed to the
            constructor of the corresponding solver type.
        """
        if n_levels < 2:
            raise ValueError("n_levels must be >= 2")

        solver_instance: VQRSolver
        if isinstance(solver, str):
            if solver not in SOLVER_TYPES:
                raise ValueError(
                    f"solver must be one of {[*SOLVER_TYPES.keys()]}, got {solver=}"
                )
            solver_opts = solver_opts or {}
            solver_instance = SOLVER_TYPES[solver](**solver_opts)
        elif isinstance(solver, VQRSolver):
            solver_instance = solver
        else:
            raise ValueError(
                f"solver must be either a string or an instance of VQRSolver"
            )

        self.solver_opts = solver_opts
        self.solver = solver_instance
        self._n_levels = n_levels
        self._fitted_solution: Optional[VQRSolution] = None

    def __sklearn_is_fitted__(self):
        return self._fitted_solution is not None

    @property
    def n_levels(self) -> int:
        """
        :return: Number of quantile levels which this estimator calculates (in every
        dimension of the target variable).
        """
        return self._n_levels

    @property
    def quantile_grid(self) -> Sequence[Array]:
        """
        :return: A sequence of quantile level grids as arrays. This is a
            d-dimensional meshgrid (see np.meshgrid) where d is the dimension of the
            target variable Y.
        """
        check_is_fitted(self)
        return self._fitted_solution.quantile_grid

    @property
    def quantile_levels(self) -> Array:
        """
        :return: An array containing the levels at which the vector quantiles were
            estimated along each target dimension.
        """
        check_is_fitted(self)
        return self._fitted_solution.quantile_levels

    @property
    def dim_y(self) -> int:
        """
        :return: The dimension of the target variable (Y).
        """
        check_is_fitted(self)
        return self._fitted_solution.dim_y

    @property
    def dim_x(self) -> int:
        """
        :return: The dimension k, of the covariates (X).
        """
        check_is_fitted(self)
        return self._fitted_solution.dim_x

    @property
    def solution_metrics(self) -> Dict[str, Any]:
        """
        :return: A dict containing solver-specific metrics about the solution.
        """
        check_is_fitted(self)
        return self._fitted_solution.metrics

    @property
    def fitted_solution(self) -> VQRSolution:
        """
        :return: The low-level VQR solution object.
        Usually it is recommended to use the high level API on this class instead.
        """
        check_is_fitted(self)
        return self._fitted_solution


class VectorQuantileEstimator(VectorQuantileBase):
    """
    Performs vector quantile estimation.
    """

    def __init__(
        self,
        n_levels: int = 50,
        solver: Union[str, VQRSolver] = DEFAULT_SOLVER_NAME,
        solver_opts: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(n_levels, solver, solver_opts)

    def vector_quantiles(self, refine: bool = False) -> QuantileFunction:
        """
        :param refine: Refine the quantile function using vector monotone rearrangement.
        :return: A QuantileFunction instance, representing a discretized version of
        the quantile function Q_{Y}(u).
        """
        check_is_fitted(self)
        qf = self._fitted_solution.vector_quantiles(X=None, refine=refine)[0]
        return qf

    def fit(self, X: Array):
        """
        Fits a quantile estimation model to the given data. In case the data
        is high-dimensional, a vector-quantile model will be fitted.
        :param X: Data of shape (n, d). Note that this is called X here to conform to
            sklearn's API. This is the target data, denoted as Y in the VQR
            problem, and we're ignoring the X in that formulation (thus making this
            estimation and not regression).
        :return: self.
        """

        # Input validation.
        N = len(X)
        Y: Array = np.reshape(X, (N, -1))

        self._fitted_solution = self.solver.solve_vqr(T=self.n_levels, Y=Y, X=None)

        return self

    def sample(self, n: int) -> Array:
        """
        Sample from Y based on the fitted vector quantile function Q(u).
        Uses the approach of Inverse transform sampling.
        https://en.wikipedia.org/wiki/Inverse_transform_sampling

        :param n: Number of samples to draw from Y|X=x.
        :return: An array containing the Sampled Y values, of shape (n, d).
        """
        check_is_fitted(self)

        # Calculate vector quantile function
        vqf = self.vector_quantiles()
        q_surfaces = tuple(vqf)  # d x (T, T, ..., T)

        # Sample from the quantile function
        return inversion_sampling(T=self.n_levels, d=self.dim_y, n=n, Qs=q_surfaces)

    def coverage(self, Y: Array, alpha: float = 0.05) -> float:
        """
        Calculates the coverage of given data points using the quantiles fitted by
        this model.

        First, it creates a contour of points in d-dimensional space which surround
        the region in which 1-2*alpha of the distribution (that was corresponds to a
        given quantile function) is contained. Then, the proportion of points
        contained within this contour is calculated.

        :param Y: Points to measure coverage for. Shape should be (N, d).
        :param alpha: Confidence level for the contour.
        :return: The coverage level, between zero and one.
        """
        check_is_fitted(self)

        # Calculate vector quantile functions
        vqf = self.vector_quantiles()
        q_surfaces = tuple(vqf)  # d x (T, T, ..., T)

        return measure_coverage(
            quantile_contour=quantile_contour(
                T=self.n_levels, d=self.dim_y, Qs=q_surfaces, alpha=alpha
            )[0],
            data=Y,
        )


class VectorQuantileRegressor(RegressorMixin, VectorQuantileBase):
    """
    Performs vector quantile regression.
    """

    def __init__(
        self,
        n_levels: int = 50,
        solver: Union[str, VQRSolver] = DEFAULT_SOLVER_NAME,
        solver_opts: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(n_levels, solver, solver_opts)
        self._scaler = StandardScaler(with_mean=True, with_std=True)

    def fit(self, X: Array, y: Array):
        """
        Fits a quantile regression model to the given data. In case the target data
        is high-dimensional, a vector-quantile model will be fitted.
        :param X: Features/covariates of shape (n, k).
        :param y: Targets/responses of shape (n, d).
        :return: self.
        """

        # Input validation.
        X, y = check_X_y(X, y, multi_output=True, ensure_2d=True)
        N = len(X)
        Y: Array = np.reshape(y, (N, -1))

        # Scale features to zero-mean
        X_scaled = self._scaler.fit_transform(X)

        self._fitted_solution = self.solver.solve_vqr(T=self.n_levels, Y=Y, X=X_scaled)

        return self

    def vector_quantiles(
        self, X: Optional[Array] = None, refine: bool = False
    ) -> Sequence[QuantileFunction]:
        """
        :param X: Covariates, of shape (N, k). Should be None if the fitted solution
        was for a VQE (un conditional quantiles).
        :param refine: Refine the conditional quantile function using vector monotone
        rearrangement.
        :return: A sequence of length N, containing QuantileFunction instances.
        Each element of the sequence corresponds to one of the covariates in X,
        and contains the discretized conditional quantile function Q_{Y|X=x}(u).
        """
        check_is_fitted(self)
        X = self._validate_X_(X, single=False)

        if X is not None:
            # Scale X with the fitted transformation before predicting
            X = self._scaler.transform(X)

        # Get the conditional quantiles
        cqfs = self._fitted_solution.vector_quantiles(X, refine=refine)
        return cqfs

    def predict(self, X: Array) -> Array:
        """
        Estimates conditional quantiles of Y|X=x.
        :param X: Samples at which to estimate. Should have shape (N, k).
        :return: An array containing the vector-quantile values.
        Returns the same as vector_quantiles, but stacked into one array.
        The result will be of shape [N, d, T, T, ..., T].
        """
        check_is_fitted(self)

        vqfs: Sequence[QuantileFunction] = self.vector_quantiles(X)

        # Stack the vector quantiles for each sample into one tensor
        vqs = np.stack(
            # Iterating over vqf produces the quantile surfaces
            [np.stack(vqf, axis=0) for vqf in vqfs],
            axis=0,
        )

        N = X.shape[0] if X is not None else 1
        assert vqs.shape == (N, self.dim_y, *[self.n_levels] * self.dim_y)
        return vqs

    def sample(self, n: int, x: Array) -> Array:
        """
        Sample from Y|X=x based on the fitted vector quantile function Q(u;x).
        Uses the approach of Inverse transform sampling.
        https://en.wikipedia.org/wiki/Inverse_transform_sampling

        :param n: Number of samples to draw from Y|X=x.
        :param x: One sample of covariates on which to condition Y.
        Should have shape  (k,) or (1, k).
        If None, then samples will be drawn from the unconditional Y.
        :return: An array containing the Sampled Y values, of shape (n, d).
        """
        check_is_fitted(self)
        x = self._validate_X_(X=x, single=True)

        # Calculate vector quantiles given sample X=x
        vqf: QuantileFunction = self.vector_quantiles(X=x)[0]
        q_surfaces = tuple(vqf)  # d x (T, T, ..., T) where each is d-dimensional

        # Sample from the quantile function
        return inversion_sampling(T=self.n_levels, d=self.dim_y, n=n, Qs=q_surfaces)

    def coverage(self, Y: Array, x: Array, alpha: float = 0.05) -> float:
        """
        Calculates the conditional coverage of given data points using the quantiles
        fitted by this model, conditioned on X=x.

        First, it creates a contour of points in d-dimensional space which surround
        the region in which 1-2*alpha of the distribution (that was corresponds to a
        given quantile function) is contained. Then, the proportion of points
        contained within this contour is calculated.

        :param Y: Points to measure coverage for. Shape should be (N, d).
        :param x: One sample of covariates on which to condition Y.
        Should have shape  (k,) or (1, k).
        :param alpha: Confidence level for the contour.
        :return: The coverage level, between zero and one.
        """
        check_is_fitted(self)
        x = self._validate_X_(X=x, single=True)

        # Calculate vector quantiles given sample X=x
        vqf: QuantileFunction = self.vector_quantiles(X=x)[0]
        q_surfaces = tuple(vqf)  # d x (T, T, ..., T) where each is d-dimensional

        return measure_coverage(
            quantile_contour=quantile_contour(
                T=self.n_levels, d=self.dim_y, Qs=q_surfaces, alpha=alpha
            )[0],
            data=Y,
        )

    def _validate_X_(self, X: Optional[Array], single: bool = False) -> Optional[Array]:
        """
        Validates the shape of a covariates array X.
        :param X: A covariates array, either (n, k) or just (k,). Can be None,
        in which case no validation happens.
        :param single: Whether to validate that n=1.
        :return: X after reshaping to (n,k).
        """
        if X is None:
            raise ValueError("Must provide covariates (X) for VQR")

        error_msg = f"X must be (k,) or ({'1' if single else 'n'}, k), got {X.shape=}"
        if np.ndim(X) == 1 and len(X) == self.dim_x:
            # reshape to (1 ,k)
            X = np.reshape(X, (1, -1))

        elif np.ndim(X) != 2:
            raise ValueError(error_msg)

        if X.shape[1] != self.dim_x:
            raise ValueError(error_msg)

        if single and X.shape[0] != 1:
            # Only a single x is supported by this method
            raise ValueError(error_msg)

        return X


class ScalarQuantileEstimator:
    def __init__(
        self,
        n_levels: int = 50,
    ):
        if n_levels < 2:
            raise ValueError("n_levels must be >= 2")

        self.n_levels = n_levels

    def fit(self, X: Array):
        N = len(X)
        Y = np.reshape(X, (N, -1))
        q = quantile(Y, q=quantile_levels(self.n_levels), axis=0)
        assert q is not None
        assert q.shape[0] == len(quantile_levels(self.n_levels))
        self._alpha = q
        return self

    @property
    def sqr_A(self) -> Array:
        return self._alpha

    @property
    def quantile_levels(self) -> Array:
        """
        :return: An array containing the levels at which the vector quantiles were
            estimated along each target dimension.
        """
        return quantile_levels(self.n_levels)

    @property
    def quantile_values(self) -> Array:
        return self.sqr_A



import logging
from typing import Optional

import click
import numpy as np
import torch
from numpy import ndarray

from vqr.api import VectorQuantileRegressor
from experiments.base import VQROptions, run_exp_context
from experiments.datasets.mvn import LinearMVNDataProvider
from experiments.utils.metrics import kde_l1, w2_keops

_LOG = logging.getLogger(__name__)


def _measure_paired_coverage(
    X: ndarray,
    Y: ndarray,
    vqr: VectorQuantileRegressor,
    n_subsample: Optional[int] = None,
    alpha: float = 0.05,
    seed: int = 42,
) -> float:
    rng = np.random.default_rng(seed)

    N, d = Y.shape
    assert len(X) == N

    if n_subsample:
        n_coverage = min(n_subsample, N)
        idx_cov = rng.permutation(n_coverage)
    else:
        idx_cov = np.arange(N)

    cov = np.mean(
        [
            # Coverage for each single data point is just 0 or 1
            vqr.coverage(y.reshape(1, d), alpha=alpha, x=x)
            for (x, y) in zip(X[idx_cov, :], Y[idx_cov, :])
        ]
    )
    return cov.item()


def single_scale_exp(
    N: int,
    T: int,
    d: int,
    k: int,
    solver_name: str,
    solver_opts: dict,
    validation_proportion: float = 0.25,
    cov_n: int = 1000,
    cov_alpha: float = 0.05,
    dist_n: int = 1000,
    seed: int = 42,
):
    # Data provider
    data_provider = LinearMVNDataProvider(d=d, k=k, seed=seed)

    # Generate data
    X, Y = data_provider.sample(
        n=int(N * (1 + validation_proportion)),
    )
    X_valid, Y_valid = X[N:, :], Y[N:, :]
    X, Y = X[:N, :], Y[:N, :]

    # Solve
    vqr = VectorQuantileRegressor(
        n_levels=T, solver=solver_name, solver_opts=solver_opts
    )
    vqr.fit(X, Y)

    # Measure coverage on n_coverage samples
    cov_train = _measure_paired_coverage(X, Y, vqr, cov_n, cov_alpha, seed)
    cov_valid = _measure_paired_coverage(X_valid, Y_valid, vqr, cov_n, cov_alpha, seed)

    # Estimate cond distribution and compare it with the gt cond distribution
    w2_dists = []
    kde_l1_dists = []
    for i in range(min(dist_n, len(Y_valid))):
        x = X_valid[[i], :]
        _, Y_gt = data_provider.sample(n=T**d, x=x)
        Y_est = vqr.sample(n=T**d, x=x)
        w2_dists.append(w2_keops(Y_gt, Y_est))
        kde_l1_dist = kde_l1(
            Y_gt,
            Y_est,
            grid_resolution=T,
            sigma=0.1,
            dtype=torch.float32,
            device="cuda" if solver_opts["gpu"] else "cpu",
        )
        kde_l1_dists.append(kde_l1_dist)

    # W2 metric
    w2_avg = np.mean(w2_dists)
    kde_l1_avg = np.mean(kde_l1_dists)

    return {
        "N": N,
        "T": T,
        "d": d,
        "k": k,
        "solver_type": solver_name,
        "solver": solver_opts,  # note: non-consistent key name on purpose
        "train_coverage": cov_train,
        "valid_coverage": cov_valid,
        "w2": w2_avg,
        "kde_l1": kde_l1_avg,
        **vqr.solution_metrics,
    }


@click.command(name="scale-exp")
@click.pass_context
@VQROptions.cli
@click.option(
    "--validation-proportion",
    type=float,
    default=0.25,
    help="Proportion of between validation set and training set.",
)
@click.option(
    "--cov-n",
    type=int,
    default=1000,
    help="Number of validation-set samples to use for coverage calculation",
)
@click.option(
    "--cov-alpha",
    type=float,
    default=0.05,
    help="Quantile level for coverage calculation.",
)
@click.option(
    "--dist-n",
    type=int,
    default=1000,
    help=(
        "Number of validation-set samples to use for conditional distribution "
        "estimation"
    ),
)
@click.option(
    "--seed",
    type=int,
    default=42,
    help="Seed for data generation",
)
def scale_exp(
    ctx: click.Context,
    validation_proportion: float,
    cov_n: int,
    cov_alpha: float,
    dist_n: int,
    seed: int,
    **kw,
):

    # Generate experiment configs from CLI
    vqr_options = VQROptions.parse_multiple(ctx)
    exp_configs = {
        vqr_option.key(): dict(
            cov_n=cov_n,
            cov_alpha=cov_alpha,
            dist_n=dist_n,
            validation_proportion=validation_proportion,
            seed=seed,
            **vqr_option.to_dict(),
        )
        for vqr_option in vqr_options
    }

    return run_exp_context(
        ctx, exp_fn=single_scale_exp, exp_configs=exp_configs, write_csv=True
    )



from typing import Tuple, Optional, Sequence

import numpy as np
from numpy.random import Generator

from experiments.datasets.base import (
    Array,
    DataProvider,
    SerializableRandomDataGenerator,
)


class LinearMVNDataProvider(DataProvider):
    """
    Generates a dataset of linearly dependent features and targets.
    Features are sampled uniformly [0, 1]. Targets are obtained via
        Y = A X + N
    where A is a random matrix whose entries are sampled i.i.d from
    standard normal distribution. N is a random vector drawn from a
    multivariate normal distribution with a random covariance matrix.

    """

    def __init__(
        self, k: int, d: int, A: Optional[Array] = None, seed: Optional[int] = 42
    ):
        """
        :param d: Dimension of targets.
        :param k: Dimension of features.
        :param seed: Random seed to use for generation. None means don't set.
        """
        super().__init__(seed=seed)
        assert d > 0
        assert k >= 0
        self._d = d
        self._k = k

        if A is not None:
            assert k == A.shape[0]
            assert d == A.shape[1]
        self._A = self._make_A() if A is None else A
        self._noise_generator = MVNNoiseGenerator(d=d, seed=self._seed, random_cov=True)

    @property
    def k(self) -> int:
        return self._k

    @property
    def d(self) -> int:
        return self._d

    def sample_x(self, n: int) -> Array:
        X = self._rng.uniform(size=(n, self.k))
        X -= np.mean(X, axis=0)
        return X

    def sample(self, n: int, x: Optional[Array] = None) -> Tuple[Array, Array]:
        if x is None:
            X = self.sample_x(n=n)
        else:
            x = np.reshape(x, (1, -1))
            assert x.shape[1] == self.k
            X = np.concatenate([x for _ in range(n)], axis=0)

        N = self._noise_generator.sample(n)
        Y = X @ self._A + N
        return X, Y

    def _make_A(self) -> Array:
        return self._rng.random(size=(self.k, self.d))


class MVNNoiseGenerator(SerializableRandomDataGenerator):
    """
    Generates multivariate normal distribution.
    """

    def __init__(
        self,
        d: int,
        seed: int = 42,
        random_cov: bool = False,
    ):
        """

        :param d: dimension of the distribution
        :param seed: Random seed to use for generation. If a random number generator
        is provided, this seed will be ignored.
        :param random_cov: Whether to use a random covariance. If set to False, a
        specific construction of covariance for anti-correlated r.v.s will be used (
        only when d=2).
        """
        assert d > 0
        super().__init__(seed=seed)
        self._d = d
        self._Sigma = self._make_Sigma(random_cov)
        self._mu = self._make_mu()

    def sample(self, n: int) -> Array:
        return self._rng.multivariate_normal(mean=self._mu, cov=self._Sigma, size=(n,))

    def _make_mu(self) -> Array:
        return np.zeros(self._d)

    def _make_Sigma(self, random_cov: bool = False) -> Array:
        if (not random_cov) and self._d == 2:
            S = np.array([[1.0, -0.7], [-0.7, 1.0]])

        else:
            # Generate random orthonormal matrix Q
            Q, R = np.linalg.qr(self._rng.normal(size=(self._d, self._d)))

            # Generate positive eigenvalues
            eigs = self._rng.uniform(size=(self._d,))

            # PSD Covariance matrix, zero mean
            S = Q.T @ np.diag(eigs) @ Q
        return S


class IndependentDataProvider(LinearMVNDataProvider):
    """
    Generates a dataset of independent features and targets.
    Features are sampled i.i.d from a standard normal distribution. In case of >1
    dimensional features, each sample in the vector-valued feature is drawn i.i.d
    from standard normal distribution.
    Targets are sampled i.i.d from a multivariate normal distribution.
    """

    def __init__(self, k: int, d: int, seed: Optional[int] = 42):
        super().__init__(k=k, d=d, A=np.zeros(shape=(k, d)), seed=seed)



from typing import Dict

from numpy import array, argsort

from vqr.api import ScalarQuantileEstimator, VectorQuantileEstimator
from vqr.coverage import measure_width, measure_coverage
from experiments.utils.split import split_train_calib_test
from experiments.datasets.mvn import IndependentDataProvider
from experiments.datasets.shapes import generate_star, generate_heart


def experiment(
    X_: array, Y_: array, alpha_low_: float, alpha_high_: float, n_levels: int
) -> Dict[str, float]:
    assert X_.shape[0] == Y_.shape[0]
    d = Y_.shape[1]

    # Split train, calibration, and test
    datasets = split_train_calib_test(X_, Y_, split_ratios=(0.0, 0.5))

    # Calibration set
    _, Y_calib = datasets["calib"]

    # Fit vector quantiles
    vq = VectorQuantileEstimator(n_levels)
    fitted_vq = vq.fit(Y_calib)

    # Get the quantiles
    VQ1, VQ2 = fitted_vq.vector_quantiles()
    T = fitted_vq.quantile_levels

    # Construct a quantile contour
    idx_low = argsort(abs(T - alpha_low_))[0]
    idx_high = argsort(abs(T - alpha_high_))[0]

    vq_contour = array(
        [
            [
                *VQ1[idx_low : idx_high + 1, idx_high + 1],
                *VQ1[idx_low, idx_low : idx_high + 1],
                *VQ1[idx_low : idx_high + 1, idx_low],
                *VQ1[idx_high + 1, idx_low : idx_high + 1],
            ],
            [
                *VQ2[idx_low : idx_high + 1, idx_high + 1],
                *VQ2[idx_low, idx_low : idx_high + 1],
                *VQ2[idx_low : idx_high + 1, idx_low],
                *VQ2[idx_high + 1, idx_low : idx_high + 1],
            ],
        ]
    ).T

    # Fit scalar quantiles
    sq = ScalarQuantileEstimator(n_levels)
    fitted_sq = sq.fit(Y_calib)
    SQ = fitted_sq.vector_quantiles()
    SQ1 = SQ[:, 0]
    SQ2 = SQ[:, 1]
    sq_contour = array(
        [
            [
                *SQ1[idx_low : idx_high + 1],
                *SQ1[idx_low : idx_high + 1],
                *[SQ1[idx_low]] * len(SQ2[idx_low : idx_high + 1]),
                *[SQ1[idx_high + 1]] * len(SQ2[idx_low : idx_high + 1]),
            ],
            [
                *[SQ2[idx_high + 1]] * len(SQ1[idx_low : idx_high + 1]),
                *[SQ2[idx_low]] * len(SQ1[idx_low : idx_high + 1]),
                *SQ2[idx_low : idx_high + 1],
                *SQ2[idx_low : idx_high + 1],
            ],
        ]
    ).T

    # Test data
    _, Y_test = datasets["test"]

    # Vector quantile: coverage and width
    vq_coverage_ = measure_coverage(vq_contour, Y_test)
    vq_width_ = measure_width(vq_contour)

    # Scalar quantile: coverage and width
    sq_coverage_ = measure_coverage(sq_contour, Y_test)
    sq_width_ = measure_width(sq_contour)

    results_ = {
        "sq_width": sq_width_,
        "sq_coverage": sq_coverage_,
        "vq_width": vq_width_,
        "vq_coverage": vq_coverage_,
        "realized_alpha_low": T[idx_low],
        "realized_alpha_high": T[idx_high],
        "deflated_alpha_low": T[idx_low],
        "deflated_alpha_high": T[idx_high],
    }

    return results_


if __name__ == "__main__":
    dataset_name = "heart"
    num_trials = 50
    n = 1500
    d_ = 2
    alpha_low = 0.05
    alpha_high = 0.95

    # Generate data
    if dataset_name == "mvn":
        X, Y = IndependentDataProvider(d=d_, k=0).sample(n=n)
    elif dataset_name == "heart":
        X, Y = generate_heart()
    elif dataset_name == "star":
        X, Y = generate_star()
    else:
        X, Y = None, None
        NotImplementedError(f"Unrecognized {dataset_name=}")

    vq_coverages = []
    vq_widths = []

    sq_coverages = []
    sq_widths = []

    for j in range(num_trials):
        results = experiment(X, Y, alpha_low, alpha_high, n_levels=40)
        vq_coverages.append(results["vq_coverage"])
        vq_widths.append(results["vq_width"])

        sq_coverages.append(results["sq_coverage"])
        sq_widths.append(results["sq_width"])

        print(
            f"Trial {j}: \n"
            f"Coverage level: [{round(results['realized_alpha_low'], 3)}, "
            f"{round(results['realized_alpha_high'], 3)}] \n"
            f"Vector quantiles: coverage={round(results['vq_coverage'], 3)}%, "
            f"width={round(results['vq_width'], 3)} \n"
            f"Deflated coverage level: [{round(results['deflated_alpha_low'], 3)}, "
            f"{round(results['deflated_alpha_high'], 3)}] \n"
            f"Scalar quantiles: coverage={round(results['sq_coverage'], 3)}%, "
            f"width={round(results['sq_width'], 3)} \n"
        )

    marginal_vq_coverage = sum(vq_coverages) / len(vq_coverages)
    average_vq_width = sum(vq_widths) / len(vq_widths)

    marginal_sq_coverage = sum(sq_coverages) / len(sq_coverages)
    average_sq_width = sum(sq_widths) / len(sq_widths)

    print(
        f"\t\t     Experiment summary \t\t \n"
        f" \t\t ========================= \t\t \n"
        f"Vector quantiles: \n"
        f"Marginal coverage over {num_trials} trials: "
        f"{marginal_vq_coverage}, average width = {average_vq_width}.\n"
        f"Scalar quantiles: \n"
        f"Marginal coverage over {num_trials} trials: "
        f"{marginal_sq_coverage}, average width = {average_sq_width}."
    )
    print(vq_coverages)
    print(vq_widths)
    print(sq_coverages)
    print(sq_widths)



from vqr.api import ScalarQuantileEstimator
from experiments.datasets.mvn import IndependentDataProvider


class TestScalarQuantileEstimator:
    def test_init(self):
        n_levels = 50
        N = 100
        d = 2
        _, Y = IndependentDataProvider(d=d, k=1).sample(n=N)

        sq = ScalarQuantileEstimator(n_levels)
        fitted_sq = sq.fit(Y)
        assert fitted_sq.quantile_values.shape[0] == n_levels
        assert len(fitted_sq.quantile_levels) == n_levels



import pickle

import ot
import numpy as np
import torch
from torch import tensor
from matplotlib import cm
from matplotlib import pyplot as plt

from vqr import VectorQuantileRegressor
from vqr.solvers.primal.cvx import SIMILARITY_FN_INNER_PROD
from experiments.utils.metrics import kde
from vqr.solvers.dual.regularized_lse import (
    RegularizedDualVQRSolver,
    MLPRegularizedDualVQRSolver,
)
from experiments.datasets.synthetic_glasses import SyntheticGlassesDataProvider

n = 5000
d = 1
k = 1
T = 100
num_epochs = 40000
linear = False
sigma = 0.1
GPU_DEVICE_NUM = 1
device = f"cuda:{GPU_DEVICE_NUM}" if GPU_DEVICE_NUM is not None else "cpu"
dtype = torch.float32
epsilon = 1e-3
X, Y = SyntheticGlassesDataProvider().sample(n)


if linear:
    solver = RegularizedDualVQRSolver(
        verbose=True,
        num_epochs=num_epochs,
        epsilon=epsilon,
        lr=2.9,
        gpu=True,
        full_precision=False,
        device_num=GPU_DEVICE_NUM,
        lr_factor=0.9,
        lr_patience=500,
        lr_threshold=0.5 * 0.01,
    )
else:
    solver = MLPRegularizedDualVQRSolver(
        verbose=True,
        num_epochs=num_epochs,
        epsilon=epsilon,
        lr=0.4,
        gpu=True,
        skip=True,
        batchnorm=True,
        hidden_layers=(1000, 1000, 1000),
        device_num=GPU_DEVICE_NUM,
        lr_factor=0.9,
        lr_patience=300,
        lr_threshold=0.5 * 0.01,
    )
vqr_est = VectorQuantileRegressor(n_levels=T, solver=solver)

vqr_est.fit(X.reshape(-1, 1), Y.reshape(-1, 1))
levels = vqr_est.quantile_levels
quantiles = []
for i in range(n):
    X_i = np.array([X[i]])[:, None]
    quantile_est = vqr_est.vector_quantiles(X=X_i)[0][0]
    cost = -SIMILARITY_FN_INNER_PROD(levels[:, None], quantile_est[None, :])
    pi = ot.emd([], [], cost)
    refined_quantiles = T * pi @ quantile_est
    quantiles.append(refined_quantiles)

quantiles = np.stack(quantiles, axis=0)
Y_est = np.concatenate(
    [vqr_est.sample(n=1, x=np.array([X[i]])[:, None])[0] for i in range(n)]
)

fig, ax = plt.subplots(1, 2, figsize=(10, 5))
ax[0].plot(X, Y, ".", label="GT Y|X")
ax[0].set_ylabel("y")
ax[0].set_xlabel("x")
ax[0].set_title("GT")
ax[0].legend()
ax[1].plot(X, Y_est, ".", label="samples of Y|X")
for quantile_level in np.linspace(0.1, 0.90, 10):
    ax[1].plot(X, quantiles[:, int(quantile_level * T)], label=f"{quantile_level:.2f}")
ax[1].legend()
ax[1].set_ylabel("y")
ax[1].set_xlabel("x")
ax[1].set_title("Non-linear QR")
plt.suptitle("Simultaneous scalar QR", fontsize="x-large")
plt.tight_layout()
plt.savefig(f"quantile_levels_{linear=}.png")

with open(f"quantiles_{linear=}.pkl", "wb") as f:
    pickle.dump({"quantiles": quantiles}, f)



import os
import queue
import logging
import multiprocessing as mp
import concurrent.futures
from time import time
from queue import Queue
from typing import (
    Any,
    Dict,
    Tuple,
    Union,
    TypeVar,
    Callable,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    cast,
)
from concurrent.futures import Future, ProcessPoolExecutor

import torch
import pandas as pd

from experiments.log import setup_logging
from experiments.utils.helpers import sec_to_time

_LOG = logging.getLogger(__name__)


_T = TypeVar("_T")


def _exp_fn_wrapper(_exp_fn, _exp_idx: int, **kwargs):
    exp_result = _exp_fn(**kwargs)
    if not exp_result or not isinstance(exp_result, dict):
        raise ValueError(
            f"Got no return value from experiment index {_exp_idx}, config={kwargs}"
        )
    return exp_result


def run_parallel_exp(
    exp_name: str,
    exp_fn: Callable[[Any], Dict[str, Any]],
    exp_configs: Dict[str, dict],
    max_workers: Optional[int] = None,
    gpu_enabled: bool = False,
    gpu_devices: Optional[Union[Sequence[int], str]] = None,
    workers_per_device: int = 1,
) -> Sequence[Dict[str, Any]]:
    """
    Runs multiple experiment configurations with parallel workers.
    :param exp_name: Name of experiment (for logging only).
    :param exp_fn: Callable that runs a single experiment configuration. It must
    return a dict with string keys.
    :param exp_configs: Mapping from a string, representing a single experiments
    id/name, to a dict of kwargs containing the configuration of that experiment.
    Each such config dict will be passed as **kwargs to exp_fn.
    :param max_workers: Maximal number of parallel worker processes to use. None
    means use number of physical cores.
    :param gpu_enabled: Whether to enable GPU support.
    :param gpu_devices: Either a list of device IDs (ints) or a comma-separated string.
    :param workers_per_device: Number of worker processes that may share a GPU.
    :return: A sequence with the results returned by each invocation of the exp_fn.
    """
    n_exps = len(exp_configs)

    if not max_workers or max_workers < 0:
        max_workers = os.cpu_count()

    if gpu_devices is None:
        # Allow all devices
        gpu_devices = tuple(range(torch.cuda.device_count()))
    elif isinstance(gpu_devices, str):
        # parse comma separated
        gpu_devices = tuple(int(d) for d in gpu_devices.split(","))

    if gpu_enabled:
        if not torch.cuda.is_available():
            _LOG.warning(f"Requested GPU, but CUDA is not available")
            gpu_enabled = False
        else:
            assert len(gpu_devices) > 0
            _LOG.info(f"GPU enabled, {gpu_devices=}")

            # Limit number of workers based on effective number of devices
            max_workers = min(max_workers, len(gpu_devices) * workers_per_device)

    _LOG.info(
        f"Starting experiment {exp_name} with {n_exps} configurations and "
        f"{max_workers} workers..."
    )
    start_time = time()

    with ProcessPoolExecutor(
        max_workers=max_workers,
        mp_context=mp.get_context("spawn"),
        initargs=cuda_worker_init_args(
            enable_cuda=gpu_enabled,
            cuda_devices=gpu_devices,
            workers_per_device=workers_per_device,
        ),
        initializer=cuda_worker_init_fn,
    ) as executor:

        futures = {
            exp_name: executor.submit(_exp_fn_wrapper, exp_fn, i, **exp_config)
            for i, (exp_name, exp_config) in enumerate(exp_configs.items())
        }

        results = []
        for i, (exp_name, exp_result) in enumerate(
            yield_future_results(
                futures,
                wait_time_sec=1.0,
                re_raise=False,
            )
        ):
            _LOG.info(
                f"Collected result {exp_name} {i+1}/{n_exps} "
                f"({100*(i+1)/n_exps:.0f}%)"
            )
            results.append(exp_result)

    elapsed_time = sec_to_time(time() - start_time)
    _LOG.info(f"Completed {len(results)}/{n_exps}, elapsed={elapsed_time}")

    return tuple(results)


def yield_future_results(
    futures: Union[Dict[_T, Future], Sequence[Future]],
    wait_time_sec=0.1,
    max_retries=None,
    re_raise=True,
    raise_on_max_retries=True,
) -> Iterator[Tuple[_T, Any]]:
    """
    Waits for futures to be ready, and yields their results. This function waits for
    each result for a fixed time, and moves to the next result if it's not ready.
    Therefore, the order of yielded results is not guaranteed to be the same as the
    order of the Future objects.

    :param futures: Either a dict mapping from some name to an
        Future to wait for, or a list of Future (in which case a name
        will be generated for each one based on it's index).
    :param wait_time_sec: Time to wait for each Future before moving to
        the next one if the current one is not ready.
    :param max_retries: Maximal number of times to wait for the same
        Future, before giving up on it. None means never give up. If
        max_retries is exceeded, an error will be logged.
    :param re_raise: Whether to re-raise an exception thrown in on of the
        tasks and stop handling. If False, exception will be logged instead and
        handling will continue.
    :param raise_on_max_retries:  Whether to raise an exception or only log an error
        in case max_retries is reached for a specific result. The type of exception
        raised will be :class:`concurrent.futures.TimeoutError`.
    :return: A generator, where each element is a tuple. The first element
        in the tuple is the name of the result, and the second element is the
        actual result. In case the task raised an exception, the second element
        will be None.
    """

    futures_d: Dict[_T, Future]
    if isinstance(futures, (list, tuple)):
        # If it's a sequence, map from index to future
        futures_d = {cast(_T, i): r for i, r in enumerate(futures)}
    elif isinstance(futures, dict):
        futures_d = futures
    else:
        raise ValueError("Expected sequence or dict of futures")

    if len(futures_d) == 0:
        raise ValueError("No futures to wait for")

    # Map result to number of retries
    retry_counts = {res_name: 0 for res_name in futures_d.keys()}

    while len(retry_counts) > 0:
        retry_counts_next = {}

        for res_name, retry_count in retry_counts.items():
            future: Future = futures_d[res_name]
            result = None

            try:
                result = future.result(timeout=wait_time_sec)

            except concurrent.futures.CancelledError:
                _LOG.warning(f"Result {res_name} was cancelled")

            except concurrent.futures.TimeoutError:
                retries = retry_counts[res_name] + 1
                if max_retries is not None and retries > max_retries:
                    msg = f"Result {res_name} timed out with {max_retries=}"
                    if raise_on_max_retries:
                        raise concurrent.futures.TimeoutError(msg)
                    _LOG.error(msg)
                else:
                    retry_counts_next[res_name] = retries

                continue

            except Exception as e:
                if re_raise:
                    raise e
                _LOG.error(f"Result {res_name} raised {type(e)}: " f"{e}", exc_info=e)

            yield res_name, result

        retry_counts = retry_counts_next


def cuda_worker_init_args(
    enable_cuda: bool = None,
    cuda_devices: Optional[Sequence[int]] = None,
    workers_per_device: int = 1,
) -> Tuple[Any, ...]:
    """
    Creates the init args which should be passed to :obj:`cuda_worker_init_fn` when it
    runs on a worker process.

    :param enable_cuda: Whether to enable CUDA on the workers. If None, the global
        CudaState will be interrogated for this value.
    :param cuda_devices: List of cuda device numbers to use.
    :param workers_per_device: Number of workers which are allowed to share a device.
    :return: A tuple containing the init args.
    """

    gpu_queue: Optional[Queue] = None

    if enable_cuda:

        if not cuda_devices:
            n_devices = torch.cuda.device_count() or 0

            # If we're using GPUs, create a queue of GPU device ids.
            # Each device id is repeated n_workers_per_device times.
            device_list = [*range(n_devices)] * workers_per_device
        else:
            device_list = [*cuda_devices] * workers_per_device

        if device_list:
            manager = mp.Manager()
            gpu_queue = manager.Queue()

            for device_id in device_list:
                gpu_queue.put(device_id)

    return (gpu_queue,)


def cuda_worker_init_fn(*args: Any) -> None:
    """
    An initializer function to be run on workers. Sets the GPU state for a worker
    process to use a worker-specific CUDA device.

    :param args: Arguments tuple which must be created by calling
        :obj:`cuda_worker_init_args`.
    """

    setup_logging()

    gpu_device_queue: Optional[Queue]
    gpu_device_queue, *_ = args

    pid = os.getpid()

    if gpu_device_queue is None:
        return

    try:
        # pull a unique device id from the queue - guarantees that different workers
        # will get different device ids (as long a unique ids were placed into the
        # queue).
        device = gpu_device_queue.get(block=False)
        torch.cuda.set_device(device)
        _LOG.info(f"Set GPU {device=} for worker process {pid}")
    except queue.Empty:
        msg = f"Can't obtain a GPU device number for worker process {pid}."
        raise ValueError(msg)



from typing import Optional, Sequence

import ot
import numpy as np
from numpy import ndarray as Array

from vqr import VQRSolution
from vqr.vqr import VQRSolver, vector_quantile_levels
from vqr.solvers.primal.cvx import SIMILARITY_FN_INNER_PROD


class POTVQESolver(VQRSolver):
    """
    Solves the VQE problem using the POT library.
    """

    def __init__(self, **pot_solver_opts):
        self._pot_solver_opts = pot_solver_opts

    @classmethod
    def solver_name(cls) -> str:
        return "vqe_pot"

    def solve_vqr(self, T: int, Y: Array, X: Optional[Array] = None) -> VQRSolution:
        # Can't deal with X's
        if X is not None:
            raise AssertionError(
                f"{self.__class__.__name__} can't work with X. It solves only "
                f"the VQE problem."
            )
        N: int = Y.shape[0]
        d: int = Y.shape[1]  # number or target dimensions

        # All quantile levels
        Td: int = T**d

        U: Array = vector_quantile_levels(T, d)
        assert U.shape == (Td, d)

        # Pairwise distances (similarity)
        S: Array = SIMILARITY_FN_INNER_PROD(U, Y.T)  # (Td, d) and (N, d)
        _, log = ot.emd2(
            M=-S,
            a=np.ones([Td]) / Td,
            b=np.ones([N]) / N,
            log=True,
            **self._pot_solver_opts,
        )

        # Obtain the lagrange multipliers Alpha (A) and Beta (B)
        A = np.reshape(-log["u"][:, None], newshape=(Td, 1))
        return VQRSolution(T, d, U, A, None)



from __future__ import annotations

import sys
import logging
from copy import deepcopy
from time import time
from typing import Any, Union, Callable, Optional, Sequence
from functools import partial

import numpy as np
import torch
from numpy import ndarray as Array
from torch import Tensor, tensor
from tqdm.auto import tqdm
from torch.optim import Optimizer
from torch.optim.lr_scheduler import ReduceLROnPlateau

from vqr import VQRSolver, VQRSolution
from vqr.vqr import vector_quantile_levels
from vqr.models import MLP

_LOG = logging.getLogger(__name__)


class RegularizedDualVQRSolver(VQRSolver):
    """
    Solves the Regularized Dual formulation of Vector Quantile Regression using
    pytorch with gradient-based optimization.

    Can solve a non-linear VQR problem, given an arbitrary neural network that acts
    as a learnable non-linear transformation of the input features.
    """

    @classmethod
    def solver_name(cls) -> str:
        return "regularized_dual"

    def __init__(
        self,
        epsilon: float = 1e-3,
        num_epochs: int = 1000,
        lr: float = 0.9,
        lr_max_steps: int = 10,
        lr_factor: float = 0.5,
        lr_patience: int = 100,
        lr_threshold: float = 5 * 0.01,
        verbose: bool = False,
        nn_init: Optional[Callable[[int], torch.nn.Module]] = None,
        batchsize_y: Optional[int] = None,
        batchsize_u: Optional[int] = None,
        inference_batch_size: int = 1,
        full_precision: bool = False,
        gpu: bool = False,
        device_num: Optional[int] = None,
        post_iter_callback: Optional[Callable[[Any], None]] = None,
    ):
        """
        :param epsilon: Regularization. The lower, the more exact the solution.
        :param num_epochs: Number of epochs (full iterations over all data) to
        optimize for.
        :param lr: Optimizer learning rate.
        :param lr_max_steps: Maximum number of times lr will be adjusted using
        :class:`ReduceLROnPlateau` scheduler. Value of 0 (default) disables scheduling.
        :param lr_factor: Factor for adjusting lr using :class:`ReduceLROnPlateau`
        scheduler. Must be in (0, 1).
        :param lr_patience: patience of :class:`ReduceLROnPlateau`.
        :param lr_threshold: threshold of :class:`ReduceLROnPlateau` (using relative
        mode).
        :param verbose: Whether to print verbose output.
        :param nn_init: Function that initializes a neural net given the number of
        input features. Must be a callable that accepts a single int (the number of
        input features, k) and returns a torch.nn.Module which for an input of
        shape (N, k) produces an output of shape (N, k').
        :param batchsize_u: The batch size of quantile levels during training.
        If set to None, full batch will be used.
        :param batchsize_y: Batch size of samples during training. If set to None,
        full batch will be used.
        :param inference_batch_size: Batch size to be used for inference. Default is 1.
        :param full_precision: Whether to use full precision (float64) or only
        double precision (float32).
        :param gpu: Whether to perform optimization on GPU. Outputs will in any case
        be numpy arrays on CPU.
        :param device_num: the GPU number on which to run, used if gpu=True. If None,
        then no device will be specified and torch will choose automatically.
        :param post_iter_callback: Optional Callback to invoke after each
        optimization iteration (after each epoch if not using batches,
        or otherwise after each batch). It should accept kwargs.
        It will be passed the following kwargs:
        (solution, batch_loss, epoch_loss, epoch_idx, batch_idx, num_epochs,
        num_batches). The solution is a VQRSolution object containing the
        intermediate solution for the iteration on which the callback is invoked.
        """
        super().__init__()

        if not 0 < epsilon < 1:
            raise ValueError(f"invalid {epsilon=}, must be in (0, 1)")
        if not num_epochs > 0:
            raise ValueError(f"invalid {num_epochs=}, must be > 0")

        if not lr > 0:
            raise ValueError(f"invalid {lr=}, must be > 0")
        if not 0 < lr_factor < 1.0:
            raise ValueError(f"invalid {lr_factor=}, must be in (0, 1)")
        if not lr_patience > 0:
            raise ValueError(f"invalid {lr_patience=}, must be > 0")
        if not 0 <= lr_threshold <= 1:
            raise ValueError(f"invalid {lr_threshold=}, must be in [0,1]")
        if not lr_max_steps >= 0:
            raise ValueError(f"invalid {lr_max_steps=}, must be >= 0")

        if (batchsize_y and not batchsize_y > 0) or (
            batchsize_u and not batchsize_u > 0
        ):
            raise ValueError(f"invalid {batchsize_y=} or {batchsize_u}, must be > 0")

        self._verbose = verbose
        self._epsilon = epsilon
        self._num_epochs = num_epochs
        self._lr = lr
        self._lr_factor = lr_factor
        self._lr_patience = lr_patience
        self._lr_threshold = lr_threshold
        self._lr_max_steps = lr_max_steps
        self._dtype = torch.float64 if full_precision else torch.float32
        self._device = (
            torch.device("cuda" if device_num is None else f"cuda:{device_num}")
            if gpu and torch.cuda.is_available()
            else torch.device("cpu")
        )
        self._dtd = dict(dtype=self._dtype, device=self._device)

        if nn_init is None:
            self._nn_init = self._default_nn_init
        else:
            self._nn_init = nn_init

        self._batchsize_y = batchsize_y
        self._batchsize_u = batchsize_u
        self._inference_batch_size = inference_batch_size
        self._callback: Optional[Callable] = post_iter_callback

    def solve_vqr(self, T: int, Y: Array, X: Optional[Array] = None) -> VQRSolution:
        start_time = time()
        log_level = logging.INFO if self._verbose else logging.NOTSET

        N = len(Y)
        Y = np.reshape(Y, (N, -1))

        if X is not None:
            X = np.reshape(X, (N, -1))

        d: int = Y.shape[1]  # number or target dimensions
        k: int = X.shape[1] if X is not None else 0  # Number of features (can be zero)

        # All quantile levels
        Td: int = T**d
        U: Array = vector_quantile_levels(T, d)
        assert U.shape == (Td, d)

        #####
        dtd = self._dtd
        epsilon = self._epsilon
        num_epochs = self._num_epochs

        Y_th = tensor(Y, **dtd)
        U_th = tensor(U, **dtd)

        psi = torch.full(size=(N, 1), fill_value=0.1, requires_grad=True, **dtd)
        X_th = None
        b = None
        net = None
        k_out = k
        if k > 0:
            X_th = tensor(X, **dtd)

            # Instantiate custom neural network
            inner_net = self._nn_init(k)
            inner_net.to(**dtd)

            # Make sure the network produces the right output shape, and use it as k
            with torch.no_grad():
                inner_net.train(False)
                net_out_shape = inner_net(torch.zeros(1, k, **dtd)).shape
            if len(net_out_shape) != 2 or net_out_shape[0] != 1:
                raise ValueError(
                    "Invalid output shape from custom neural net: "
                    "expected (N, k_in) input to produce (N, k_out) output"
                )
            k_out = net_out_shape[1]

            # Add a non-trainable BatchNorm at the end
            net = torch.nn.Sequential(
                inner_net,
                torch.nn.BatchNorm1d(
                    num_features=k_out, affine=False, track_running_stats=True, **dtd
                ),
            )
            # drop reference to inner_net to prevent memory being held after we move
            # net to RAM
            del inner_net
            net.train(True)  # note: also applied to inner_net
            b = torch.zeros(*(Td, k_out), requires_grad=True, **dtd)

        optimizer = torch.optim.SGD(
            [
                dict(params=[*net.parameters()] if k > 0 else []),
                dict(
                    params=[b, psi] if k > 0 else [psi],
                ),
            ],
            lr=self._lr,
            momentum=0.9,
            nesterov=True,
            weight_decay=0.0,
        )

        scheduler = ReduceLROnPlateau(
            optimizer=optimizer,
            mode="min",
            factor=self._lr_factor,
            # loss needs to decrease by factor of 1-threshold every patience epochs
            # in min+rel mode: dynamic_threshold = best_loss * ( 1 - threshold )
            patience=self._lr_patience,
            threshold=self._lr_threshold,
            threshold_mode="rel",
            # when lr_max_steps==0, min_lr==lr so it will not be reduced.
            min_lr=self._lr * self._lr_factor**self._lr_max_steps,
            verbose=False,
        )

        _LOG.log(
            log_level,
            f"{self}: Solving with {N=}, {T=}, {d=}, {k=}, {k_out=}, "
            f"batchsize_y={self._batchsize_y}, batchsize_u={self._batchsize_u}",
        )

        final_loss = None
        with tqdm(
            total=num_epochs,
            file=sys.stdout,
            unit="epochs",
            disable=not self._verbose,
            mininterval=0.2,
            position=0,
            leave=True,
            ncols=100,
        ) as pbar:
            if self._batchsize_y or self._batchsize_u:
                final_loss = self._train_minibatch(
                    T, Y_th, U_th, psi, X_th, b, net, optimizer, scheduler, pbar
                )

            else:
                final_loss = self._train_fullbatch(
                    T, Y_th, U_th, psi, X_th, b, net, optimizer, scheduler, pbar
                )

        train_time = time() - start_time

        # Move data back to main memory, and free GPU memory for inference
        Y_th = Y_th.cpu()
        U_th = U_th.cpu()
        psi = psi.detach_().cpu()
        if k > 0:
            X_th = X_th.cpu()
            b = b.detach_().cpu()
            net = net.cpu()
            net.train(False)  # Finalize network
        torch.cuda.empty_cache()

        # Finalize phi and calculate VQR coefficients A and B
        phi = self._evaluate_phi_inference(Y_th, U_th, psi, epsilon, X_th, b, net)

        total_time = time() - start_time
        inference_time = total_time - train_time
        _LOG.log(log_level, f"{self}: total_time={total_time:.2f}s")

        return self._create_solution(
            T=T,
            d=d,
            k=k,
            U=U,
            phi=phi,
            b=b,
            net=net,
            solution_metrics=dict(
                train_time=train_time,
                inference_time=inference_time,
                total_time=total_time,
                final_loss=final_loss,
            ),
        )

    def _create_solution(
        self, T, d, k, U, phi, b, net, solution_metrics: dict = None
    ) -> VQRSolution:

        A = phi.detach().cpu().numpy()
        B = None
        x_transform_fn = None
        if k > 0:
            B = b.detach().cpu().numpy()

            # Create a copy of the net on CPU, without moving the original net to CPU
            # Note: deepcopy seems to create a copy of all tensors on cpu, but calling
            # cpu() to make sure.
            net_copy: torch.nn.Module = deepcopy(net).cpu()
            x_transform_fn = partial(
                self._features_transform, net=net_copy, dtype=self._dtype
            )

        return VQRSolution(
            T,
            d,
            U,
            A,
            B,
            X_transform=x_transform_fn,
            k_in=k,
            solution_metrics=solution_metrics,
        )

    @staticmethod
    def _features_transform(
        X: Array, net: torch.nn.Module, dtype: torch.dtype
    ) -> Array:
        # Assumes net is on cpu and in eval mode.
        X_th = torch.from_numpy(X).to(dtype=dtype)
        with torch.no_grad():
            return net(X_th).numpy()

    def _train_fullbatch(
        self,
        T: int,
        Y: Tensor,
        U: Tensor,
        psi: Tensor,
        X: Tensor,
        b: Tensor,
        net: torch.nn.Module,
        optimizer: Optimizer,
        scheduler: Any,
        pbar: tqdm,
    ) -> float:
        N = len(Y)
        Td = len(U)
        d: int = Y.shape[1]
        k: int = X.shape[1] if X is not None else 0

        mu = torch.ones(Td, 1, **self._dtd) / Td
        nu = torch.ones(N, 1, **self._dtd) / N

        UY = U @ Y.T  # (T^d, N)

        epsilon = self._epsilon
        objective = tensor(float("nan"))
        for epoch_idx in range(self._num_epochs):

            # Optimize
            optimizer.zero_grad()
            phi = self._evaluate_phi(Y, U, psi, epsilon, X, b, net, UY)
            constraint_loss = phi.T @ mu
            objective = psi.T @ nu + constraint_loss
            objective.backward()
            optimizer.step()
            scheduler.step(objective)

            # Update progress and stats
            pbar.update(1)
            pbar.set_postfix(
                total_loss=objective.item(),
                constraint_loss=constraint_loss.item(),
                lr=optimizer.param_groups[0]["lr"],
                refresh=False,
            )

            # Invoke callback
            if self._callback:
                self._callback(
                    solution=self._create_solution(T, d, k, U, phi, b, net),
                    batch_loss=None,
                    epoch_loss=objective.item(),
                    epoch_idx=epoch_idx,
                    batch_idx=None,
                    num_epochs=self._num_epochs,
                    num_batches=None,
                    xy_slice=None,
                    u_slice=None,
                )

        return objective.item()

    def _train_minibatch(
        self,
        T: int,
        Y: Tensor,
        U: Tensor,
        psi: Tensor,
        X: Tensor,
        b: Tensor,
        net: torch.nn.Module,
        optimizer: Optimizer,
        scheduler: Any,
        pbar: tqdm,
    ) -> float:
        N = len(Y)
        Td = len(U)
        d: int = Y.shape[1]
        k: int = X.shape[1] if X is not None else 0

        epsilon = self._epsilon

        def _num_batches(num_samples, batch_size):
            if not batch_size:
                batch_size = num_samples
            return int(np.ceil(num_samples / batch_size))

        def _yield_batches(num_samples, batch_size):
            if not batch_size or batch_size > num_samples:
                batch_size = num_samples

            while True:
                idx = np.concatenate(
                    [
                        np.random.permutation(num_samples),
                        np.random.permutation(num_samples),
                    ],
                    axis=0,
                )

                for batch_idx in range(_num_batches(num_samples, batch_size)):
                    batch_slice = idx[
                        (batch_size * batch_idx) : (batch_size * (batch_idx + 1))
                    ]

                    if batch_size == num_samples:
                        batch_slice = sorted(batch_slice)

                    assert len(batch_slice) == batch_size
                    yield batch_slice

        num_batches_xy = _num_batches(N, self._batchsize_y)
        num_batches_u = _num_batches(Td, self._batchsize_u)
        num_batches_per_epoch = max(num_batches_xy, num_batches_u)

        total_objective = tensor(float("nan"))
        for epoch_idx in range(self._num_epochs):

            total_objective = tensor(0.0, **self._dtd)

            for batch_idx, xy_slice, u_slice in zip(
                range(num_batches_per_epoch),
                _yield_batches(N, self._batchsize_y),
                _yield_batches(Td, self._batchsize_u),
            ):
                Y_batch = Y[xy_slice]
                U_batch = U[u_slice]
                psi_batch = psi[xy_slice, :]
                mu_batch = torch.ones(len(U_batch), 1, **self._dtd) / len(U_batch)
                nu_batch = torch.ones(len(Y_batch), 1, **self._dtd) / len(Y_batch)
                X_batch, b_batch = None, None
                if X is not None:
                    X_batch = X[xy_slice]
                    b_batch = b[u_slice, :]

                # Optimize
                optimizer.zero_grad()
                phi_batch = self._evaluate_phi(
                    Y_batch, U_batch, psi_batch, epsilon, X_batch, b_batch, net, UY=None
                )
                constraint_loss = phi_batch.T @ mu_batch
                objective = psi_batch.T @ nu_batch + constraint_loss
                objective.backward()
                optimizer.step()

                total_objective = total_objective + objective

                # Invoke callback
                if self._callback:
                    if not self._batchsize_u:
                        phi_all_levels = phi_batch
                    else:
                        # In the case of U-batches, we need to compute phi with all
                        # T^d levels, otherwise we can't create a valid solution object
                        with torch.no_grad():
                            phi_all_levels = self._evaluate_phi(
                                Y_batch, U, psi_batch, epsilon, X_batch, b, net, UY=None
                            )
                    self._callback(
                        solution=self._create_solution(
                            T, d, k, U, phi_all_levels, b, net
                        ),
                        batch_loss=objective.item(),
                        epoch_loss=total_objective.item() / (batch_idx + 1),
                        epoch_idx=epoch_idx,
                        batch_idx=batch_idx + epoch_idx * num_batches_per_epoch,
                        num_epochs=self._num_epochs,
                        num_batches=num_batches_per_epoch * self._num_epochs,
                        xy_slice=xy_slice,
                        u_slice=u_slice,
                    )

            total_objective /= num_batches_per_epoch
            scheduler.step(total_objective)

            # Update progress and stats
            pbar.update(1)
            pbar.set_postfix(
                total_loss=total_objective.item(),
                lr=optimizer.param_groups[0]["lr"],
                batch=(
                    f"{batch_idx + epoch_idx * num_batches_per_epoch}/"
                    f"{num_batches_per_epoch * self._num_epochs}"
                ),
                refresh=False,
            )

        return total_objective.item()

    @staticmethod
    def _evaluate_phi(
        Y: Tensor,
        U: Tensor,
        psi: Tensor,
        epsilon: float,
        X: Tensor,
        b: Tensor,
        net: torch.nn.Module,
        UY: Optional[Tensor] = None,
    ):
        """
        Calculates the phi optimization variable of the VQR objective.

        Phi is defined for quantile level i as follows:
            phi_i = max_j { u_i y_j - b_i x_j - psi_j }

        This implementation uses a log-sum-exp reduction with
        epsilon-smoothing instead of taking max_j.
        """
        if UY is None:
            UY = U @ Y.T  # (T^d, N)

        bX = 0
        if X is not None:
            bX = b @ net(X).T

        max_arg = UY - bX - psi.reshape(1, -1)  # (T^d,N)-(T^d,N)-(1,N) = (T^d, N)
        phi = epsilon * torch.logsumexp(max_arg / epsilon, dim=1)

        return phi.reshape(-1, 1)  # (T^d, 1)

    def _evaluate_phi_inference(
        self,
        Y: Tensor,
        U: Tensor,
        psi: Tensor,
        epsilon: float,
        X: Tensor,
        b: Tensor,
        net: torch.nn.Module,
    ):
        Td, d = U.shape
        N, _ = Y.shape

        phi_full = torch.empty(Td, 1, dtype=self._dtype, device="cpu")
        num_batches_us = int(np.ceil(Td / self._inference_batch_size))
        idx = np.arange(Td)
        Y = Y.to(self._device)
        psi = psi.to(self._device)
        X = X.to(self._device) if X is not None else None
        net = net.to(self._device) if net is not None else None

        with torch.no_grad():
            for batch_idx in range(num_batches_us):
                batch_slice = idx[
                    self._inference_batch_size
                    * batch_idx : min(self._inference_batch_size * (batch_idx + 1), Td)
                ]
                U_batch = U[batch_slice].to(self._device)
                b_batch = b[batch_slice].to(self._device) if b is not None else None
                phi = self._evaluate_phi(
                    Y,
                    U_batch,
                    psi,
                    epsilon,
                    X,
                    b_batch,
                    net,
                    UY=None,
                )
                phi = phi.cpu().detach()
                phi_full[batch_slice] = phi

        return phi_full.reshape(-1, 1)

    def _default_nn_init(self, *args, **kwargs):
        """
        Default neural-net init function. For internal use.
        """
        return torch.nn.Identity()

    def __getstate__(self):
        # Prevent serialization of callback if the solver is serialized.
        state = self.__dict__.copy()
        state["_callback"] = None
        return state

    def __repr__(self):
        return f"{self.__class__.__name__}(eps={self._epsilon:.0e})"


class MLPRegularizedDualVQRSolver(RegularizedDualVQRSolver):
    """
    Same as RegularizedDualVQRSolver, but with a general-purpose MLP as a
    learnable non-linear feature transformation.
    """

    @classmethod
    def solver_name(cls) -> str:
        return "regularized_dual_mlp"

    def __init__(
        self,
        hidden_layers: Union[str, Sequence[int]] = (32,),
        activation: Union[str, torch.nn.Module] = "relu",
        skip: bool = True,
        batchnorm: bool = False,
        dropout: float = 0,
        **solver_opts,
    ):
        """
        Supports init args of both :class:`MLP` and :class:`RegularizedDualVQRSolver`.
        """
        if "nn_init" in solver_opts:
            raise ValueError("Can't provide nn_init to this solver")

        super().__init__(
            nn_init=partial(
                MLP,
                hidden_dims=hidden_layers,
                nl=activation,
                skip=skip,
                batchnorm=batchnorm,
                dropout=dropout,
            ),
            **solver_opts,
        )



import os
from pathlib import Path

EXPERIMENTS_ROOT_DIR = Path(__file__).parent.resolve()
PROJECT_ROOT_DIR = EXPERIMENTS_ROOT_DIR.parent.resolve()

EXPERIMENTS_DATA_DIR = PROJECT_ROOT_DIR.joinpath("data")

EXPERIMENTS_OUT_DIR = EXPERIMENTS_ROOT_DIR.parent.joinpath("out")
os.makedirs(EXPERIMENTS_OUT_DIR, exist_ok=True)

# When inside a conda env, prioritize lib and header dirs from the env
conda_prefix = os.environ.get("CONDA_PREFIX")
if conda_prefix:
    # Fix CPATH for KeOps
    cpath = os.environ.get("CPATH", "")
    os.environ["CPATH"] = f"{conda_prefix}/include:{cpath}"

    # Fix import errors with torch when global cuda toolkit is installed
    ld_lib_path = os.environ.get("LD_LIBRARY_PATH", "")
    os.environ["LD_LIBRARY_PATH"] = f"{conda_prefix}/lib64:{ld_lib_path}"



import numpy as np
import pytest

from experiments.utils.metrics import kde_l1, w2_pot, w2_keops


class TestMetrics:
    def _generate_data(self, N: int, d: int):
        dtype = np.float32

        # Y1 & Y2 are from the same dist, Y3 is not
        Y1 = np.random.randn(N, d).astype(dtype)
        Y2 = np.random.randn(N, d).astype(dtype)
        Y3 = np.random.randn(N, d).astype(dtype) + 1.0 + (d - 1) * 0.1

        return Y1, Y2, Y3

    @pytest.mark.parametrize("resolution", [10, 12])
    @pytest.mark.parametrize("N", [1000, 2000, 3000])
    @pytest.mark.parametrize("d", [1, 2, 3, 4])
    def test_kde_l1(self, N, d, resolution):
        Y1, Y2, Y3 = self._generate_data(N, d)
        kde_same_dist = kde_l1(
            Y1, Y2, device="cpu", grid_resolution=resolution, sigma=0.05
        )
        kde_diff_dist = kde_l1(
            Y1, Y3, device="cpu", grid_resolution=resolution, sigma=0.05
        )
        print(f"{N=}, {d=}, {resolution=}, {kde_same_dist=:.3e}, {kde_diff_dist=:.3e}")
        assert kde_same_dist < kde_diff_dist

    @pytest.mark.parametrize("N", [1000, 2000, 3000])
    @pytest.mark.parametrize("d", [1, 2, 3, 4])
    def test_w2_keops(self, N, d):
        Y1, Y2, Y3 = self._generate_data(N, d)
        w2_same_dist = w2_keops(Y1, Y2, device="cpu")
        w2_diff_dist = w2_keops(Y1, Y3, device="cpu")
        print(f"{N=}, {d=}, {w2_same_dist=:.3e}, {w2_diff_dist=:.3e}")
        assert w2_same_dist < w2_diff_dist

    @pytest.mark.parametrize("N", [1000, 2000, 3000])
    @pytest.mark.parametrize("d", [1, 2, 3, 4])
    def test_w2_pot(self, N, d):

        Y1, Y2, Y3 = self._generate_data(N, d)
        w2_same_dist = w2_pot(Y1, Y2)
        w2_diff_dist = w2_pot(Y1, Y3)
        print(f"{N=}, {d=}, {w2_same_dist=:.3e}, {w2_diff_dist=:.3e}")
        assert w2_same_dist < w2_diff_dist



from typing import Tuple, Optional, Sequence

import numpy as np
from numpy import ndarray
from numpy import ndarray as Array
from matplotlib import pyplot as plt
from matplotlib.cm import ScalarMappable
from scipy.spatial import ConvexHull
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from mpl_toolkits.mplot3d import Axes3D

from vqr.vqr import quantile_levels, quantile_contour
from vqr.coverage import point_in_hull


def _level_label(t: float) -> str:
    """
    Generates labels for quantile levels in plots
    :param t: quantile level
    :return: plot label
    """
    s = f"{t:.2f}"
    if s.startswith("0"):
        s = s[1:]
    return s


def plot_quantiles(
    T: int,
    d: int,
    Qs: Sequence[Array],
    Us: Sequence[Array],
    figsize: Optional[Tuple[int, int]] = None,
) -> Figure:
    """
    Plots scalar quantiles (d=1) or vector quantiles of d=2 obtained from the
    solution of the VQR optimal transport problem.

    A new figure will be created. Scalar quantiles will be
    plotted using a simple line plot, while vector quantiles with d=2 will be plotted
    as an image, where the pixel colors correspond to quantile value.

    :param T: The number of quantile levels that was used for solving the problem.
    :param d: The dimension of the target data (Y) that was used for solving the
    problem.
    :param Qs: Quantile surfaces per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :param Us: Quantile levels per dimension of U. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :param figsize: Size of figure to create. Will be passed to plt.subplots.
    :return: The created figure.
    """
    if d > 2:
        raise RuntimeError("Can't plot quantiles with dimension greater than 2")

    assert d == len(Qs) == len(Us)

    fig: Figure
    _axes: ndarray
    fig, _axes = plt.subplots(
        nrows=1,
        ncols=d,
        figsize=figsize,
        squeeze=False,
    )
    # _axes is (1, d), take first row to get (d,)
    axes: Sequence[Axes] = list(_axes[0])

    levels: ndarray = quantile_levels(T)
    tick_labels = [_level_label(t) for t in levels]

    for i, (ax, Q) in enumerate(zip(axes, Qs)):
        if d == 1:
            ax.plot(*Us, Q)
            ax.set_title(f"$Q_{{{i + 1}}}(u_1)$")
            ax.set_xlabel(f"$u_1$")
            ax.set_xticks(levels)
            ax.set_xticklabels(
                tick_labels, rotation=90, ha="right", rotation_mode="anchor"
            )
            ax.grid(True)

        elif d == 2:
            Q = Q.T  # to match the axes
            m = ax.imshow(Q, aspect="equal", interpolation="none", origin="lower")

            ticks = levels * T - 1
            ax.set_title(f"$Q_{{{i + 1}}}(u_1, u_2)$")
            ax.set_yticks(ticks)
            ax.set_yticklabels(tick_labels)
            ax.set_ylabel("$u_1$")
            ax.xaxis.set_ticks_position("bottom")
            ax.set_xticks(ticks)
            ax.set_xticklabels(
                tick_labels, rotation=90, ha="right", rotation_mode="anchor"
            )
            ax.set_xlabel("$u_2$")

            fig.colorbar(m, ax=[ax], shrink=0.2)

        ax.locator_params(axis="both", tight=True, nbins=20)
    return fig


def plot_quantiles_3d(
    T,
    d,
    Qs: Sequence[Array],
    Us: Sequence[Array],
    colorbar: bool = True,
    cmap: str = "viridis",
    alpha: float = 1.0,
    figsize: Optional[Tuple[int, int]] = None,
) -> Tuple[Figure, Sequence[Axes]]:
    """
    Plots vector quantiles of d=2 or d=3 obtained from the solution of the VQR optimal
    transport problem.

    A new figure will be created. A three-dimensional plot will be created, where for d=2
    the quantiles will be plotted as surfaces, and for d=3 the quantiles will be
    plotted as voxels, where the color of the quantile corresponds to the value of the
    quantile.

    :param T: The number of quantile levels that was used for solving the problem.
    :param d: The dimension of the target data (Y) that was used for solving the
        problem.
    :param Qs: Quantile surfaces per dimension of Y. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :param Us: Quantile levels per dimension of U. A sequence of length d,
    where each element is of shape (T, T, ..., T).
    :param colorbar: whether to add a colorbar.
    :param cmap: Colormap name.
    :param alpha: Color alpha value (transparency).
    :param figsize: Size of figure to create. Will be passed to plt.subplots.
    :return: The created figure.
    """
    if not 1 < d < 4:
        raise RuntimeError("Can't plot 3d quantiles with dimension other than 2, 3")

    # Transpose to match axes
    # TODO: Need to check why this is needed
    axes_perm = [*range(1, d), 0]
    Qs = [np.transpose(Q, axes_perm) for Q in Qs]
    Us = [np.transpose(U, axes_perm) for U in Us]

    fig: Figure
    _axes: ndarray
    fig, _axes = plt.subplots(
        nrows=1,
        ncols=d,
        figsize=figsize,
        squeeze=False,
        subplot_kw={"projection": "3d"},
    )
    axes: Sequence[Axes3D] = list(_axes[0])

    levels: ndarray = quantile_levels(T)
    tick_labels = [_level_label(t) for t in levels]

    for i, (ax, Q) in enumerate(zip(axes, Qs)):
        if d == 2:
            ticks = levels
            m = ax.plot_surface(*Us, Q, cmap=cmap, alpha=alpha)
            ax.set_title(f"$Q_{{{i + 1}}}(u_1, u_2)$")
            ax.set_xlabel("$u_1$")
            ax.set_ylabel("$u_2$")
            if colorbar:
                fig.colorbar(m, ax=[ax], shrink=0.2)

        if d == 3:
            ticks = levels * T - 1
            cm = plt.get_cmap(cmap)
            norm = plt.Normalize(np.min(Q), np.max(Q))
            active_voxesls = np.ones_like(Q)
            ax.voxels(
                active_voxesls,
                facecolors=cm(norm(Q)),
                edgecolors=None,
                alpha=alpha,
            )
            ax.set_zticks(ticks)
            ax.set_zticklabels(tick_labels)
            ax.set_title(f"$Q_{{{i + 1}}}(u_1, u_2, u_3)$")
            ax.set_xlabel("$u_1$")
            ax.set_ylabel("$u_2$")
            ax.set_zlabel("$u_3$")

            if colorbar:
                fig.colorbar(ScalarMappable(norm=norm, cmap=cm), ax=[ax], shrink=0.2)

        ax.set_yticks(ticks)
        ax.set_yticklabels(tick_labels)
        ax.set_xticks(ticks)
        ax.set_xticklabels(tick_labels)

        ax.locator_params(axis="both", tight=True, nbins=10)

    return fig, axes


def plot_contour_2d(
    Qs: Sequence[Array],
    Y: Array = None,
    alphas: Sequence[float] = (0.1,),
    alpha_labels: Sequence[str] = None,
    alpha_colors: Sequence[str] = None,
    ax: Axes = None,
):
    """
    Plots two-dimensional coverage plots.
    Can show both training and validation data, and calculate coverage.

    :param Qs: The quantile surfaces
    :param Y: optional data to plot
    :param alphas: Contour levels to plot.
    :param ax: Optional axes object to plot to. If not provided, will create a new
    figure.
    :return: A tuple with the (Figure, Axis).
    """
    assert len(Qs) == 2
    Q1, Q2 = Qs

    T = Q1.shape[0]
    assert Q1.shape == Q2.shape == (T, T)
    assert all(0.0 < alpha < 0.5 for alpha in alphas)

    if alpha_labels:
        assert len(alpha_labels) == len(alphas)
    if alpha_colors:
        assert len(alpha_colors) == len(alphas)

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(8, 8))
    else:
        fig, ax = None, ax

    if Y is not None:
        assert np.ndim(Y) == 2 and Y.shape[1] == 2
        ax.scatter(
            Y[:, 0],
            Y[:, 1],
            alpha=0.4,
            color="k",
            marker=".",
            edgecolor="white",
            label="Y",
        )

    for i, alpha in enumerate(alphas):
        Q_contour = quantile_contour(T, d=2, Qs=[Q1, Q2], alpha=alpha)[0]

        label = alpha_labels[i] if alpha_labels else rf"$\alpha$={alpha:.2f}"
        color = alpha_colors[i] if alpha_colors else f"C{i}"
        surf_kws = dict(alpha=0.5, color=color, s=200, marker=".", label=label)
        ax.scatter(
            *Q_contour.T,
            **surf_kws,
        )

    ax.legend()
    return fig, ax


def plot_coverage_2d(
    Q1: Array,
    Q2: Array,
    Y_valid: Array = None,
    Y_train: Array = None,
    alpha: float = 0.1,
    ax: Axes = None,
    title: str = None,
    xylim: str = None,
    xlabel: str = None,
    ylabel: str = None,
    contour_color: str = "k",
    contour_label: str = None,
):
    """
    Plots two-dimensional coverage plots.
    Can show both training and validation data, and calculate coverage.

    :param Q1: The first 2d quantile surface.
    :param Q2: The second 2d quantile surface.
    :param Y_valid: Validation data points (N, 2).
    :param Y_train: Training data points (N', 2)
    :param alpha: Quantile level for coverage calculation. Should be <0.5.
    E.g., 0.05 means the contour will correspond to 0.95 of the data in each dimension.
    :param ax: Optional axes object to plot to. If not provided, will create a new
    figure.
    :param title: Title to give the axes.
    :param xylim: Limits for x and y axes.
    :param xlabel: Label for x axis.
    :param ylabel: Label for y axis.
    :param contour_color: Color for contour line.
    :param contour_label: Legend label for contour line.
    :return: A tuple with the (Figure, Axis, validation coverage).
    """
    T = Q1.shape[0]
    assert Q1.shape == Q2.shape == (T, T)
    assert 0.0 < alpha < 0.5

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(8, 8))
    else:
        fig, ax = None, ax

    Q_surface = quantile_contour(T, d=2, Qs=[Q1, Q2], alpha=alpha)[0]

    surf_kws = dict(alpha=0.5, color=contour_color, s=200, marker="v")
    ax.scatter(*Q_surface.T, **surf_kws)

    # Plot convex hull
    hull = ConvexHull(Q_surface)
    for i, simplex in enumerate(hull.simplices):
        label = None
        if i == len(hull.simplices) - 1:
            label = contour_label or rf"Quantile Contour ($\alpha$={alpha})"
        ax.plot(
            Q_surface[simplex, 0],
            Q_surface[simplex, 1],
            color=contour_color,
            label=label,
        )

    # Plot training data
    train_coverage = None
    if Y_train is not None:
        is_in_hull = np.array([point_in_hull(p, hull) for p in Y_train])
        train_coverage = np.mean(is_in_hull).item()
        ax.scatter(
            *Y_train.T,
            color="k",
            alpha=0.3,
            label=f"training data (cov={train_coverage * 100:.2f})",
        )

    val_coverage = None
    # Plot validation data
    if Y_valid is not None:
        is_in_hull = np.array([point_in_hull(p, hull) for p in Y_valid])
        val_coverage = np.mean(is_in_hull).item()
        ax.scatter(
            *Y_valid[is_in_hull, :].T,
            marker="x",
            color="g",
            label=f"validation (cov={val_coverage * 100:.2f})",
        )
        ax.scatter(
            *Y_valid[~is_in_hull, :].T,
            marker="x",
            color="m",
            label="validation outliers",
        )

    ax.set_xlim(xylim)
    ax.set_ylim(xylim)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)
    ax.legend()
    return fig, ax, val_coverage
